# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("to_torch",)

np = NumpyMetadata.instance()


@high_level_function()
def to_torch(array):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).

    Converts `array` (only ListOffsetArray, ListArray, RegularArray and NumpyArray data types supported)
    into a PyTorch Tensor, if possible.

    If `array` contains any other data types (RecordArray for example) the function raises a TypeError.
    """

    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array)


def _impl(array):
    try:
        import torch
    except ImportError as err:
        raise ImportError(
            """to use ak.to_torch, you must install 'torch' package with:

         pip install torch

or

        conda install pytorch"""
        ) from err

    # useful function that handles all possible input arrays
    array = ak.to_layout(array, allow_record=False)

    # get the device array is on
    device = ak.backend(array)

    if device not in ["cuda", "cpu"]:
        raise ValueError("Only 'cpu' and 'cuda' backend conversions are allowed")

    # convert to numpy or cupy if `array` on gpu
    try:
        backend_array = array.to_backend_array(allow_missing=False)
    except ValueError as err:
        raise TypeError(
            "Only arrays containing equal-length lists of numbers can be converted into a PyTorch Tensor"
        ) from err

    # check if cupy or numpy
    if isinstance(backend_array, np.ndarray):
        # convert numpy to a torch tensor
        tensor = torch.from_numpy(backend_array)
    else:
        # cupy -> torch tensor
        tensor = torch.utils.dlpack.from_dlpack(backend_array.toDlpack())

    return tensor



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import copy
from collections.abc import Mapping, MutableMapping, Sequence

import awkward as ak
from awkward._backends.backend import Backend
from awkward._layout import maybe_posaxis
from awkward._meta.indexedoptionmeta import IndexedOptionMeta
from awkward._nplikes.array_like import ArrayLike, maybe_materialize
from awkward._nplikes.jax import Jax
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpy_like import IndexType, NumpyMetadata
from awkward._nplikes.placeholder import PlaceholderArray
from awkward._nplikes.shape import ShapeItem, unknown_length
from awkward._nplikes.typetracer import MaybeNone, TypeTracer
from awkward._nplikes.virtual import VirtualNDArray
from awkward._parameters import (
    parameters_intersect,
    parameters_union,
)
from awkward._regularize import is_integer_like
from awkward._slicing import NO_HEAD
from awkward._typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Final,
    Literal,
    Self,
    SupportsIndex,
    final,
)
from awkward._util import UNSET
from awkward.contents.content import (
    ApplyActionOptions,
    Content,
    ImplementsApplyAction,
    RemoveStructureOptions,
    ToArrowOptions,
)
from awkward.errors import AxisError
from awkward.forms.form import Form, FormKeyPathT
from awkward.forms.indexedoptionform import IndexedOptionForm
from awkward.index import Index

if TYPE_CHECKING:
    from awkward._slicing import SliceItem

np = NumpyMetadata.instance()
numpy = Numpy.instance()


@final
class IndexedOptionArray(IndexedOptionMeta[Content], Content):
    """
    IndexedOptionArray is an #ak.contents.IndexedArray for which
    negative values in the index are interpreted as missing. It represents
    #ak.types.OptionType data like #ak.contents.ByteMaskedArray,
    #ak.contents.BitMaskedArray, and #ak.contents.UnmaskedArray, but
    the flexibility of the arbitrary `index` makes it a common output of
    many operations.

    IndexedOptionArray doesn't have a direct equivalent in Apache Arrow.

    To illustrate how the constructor arguments are interpreted, the following is a
    simplified implementation of `__init__`, `__len__`, and `__getitem__`:

        class IndexedOptionArray(Content):
            def __init__(self, index, content):
                assert isinstance(index, (Index32, Index64))
                assert isinstance(content, Content)
                for x in index:
                    assert x < len(content)  # index[i] may be negative
                self.index = index
                self.content = content

            def __len__(self):
                return len(self.index)

            def __getitem__(self, where):
                if isinstance(where, int):
                    if where < 0:
                        where += len(self)
                    assert 0 <= where < len(self)
                    if self.index[where] < 0:
                        return None
                    else:
                        return self.content[self.index[where]]

                elif isinstance(where, slice) and where.step is None:
                    return IndexedOptionArray(
                        self.index[where.start : where.stop], self.content
                    )

                elif isinstance(where, str):
                    return IndexedOptionArray(self.index, self.content[where])

                else:
                    raise AssertionError(where)
    """

    def __init__(self, index, content, *, parameters=None):
        if not (
            isinstance(index, Index)
            and index.dtype
            in (
                np.dtype(np.int32),
                np.dtype(np.int64),
            )
        ):
            raise TypeError(
                f"{type(self).__name__} 'index' must be an Index with dtype in (int32, uint32, int64), "
                f"not {index!r}"
            )
        if not isinstance(content, Content):
            raise TypeError(
                f"{type(self).__name__} 'content' must be a Content subtype, not {content!r}"
            )
        is_cat = parameters is not None and parameters.get("__array__") == "categorical"
        if (content.is_union and not is_cat) or content.is_indexed or content.is_option:
            raise TypeError(
                "{0} cannot contain a union-type (unless categorical), option-type, or indexed 'content' ({1}); try {0}.simplified instead".format(
                    type(self).__name__, type(content).__name__
                )
            )

        assert index.nplike is content.backend.nplike

        self._index = index
        self._content = content
        self._init(parameters, content.backend)

    @property
    def index(self):
        return self._index

    form_cls: Final = IndexedOptionForm

    def copy(self, index=UNSET, content=UNSET, *, parameters=UNSET):
        return IndexedOptionArray(
            self._index if index is UNSET else index,
            self._content if content is UNSET else content,
            parameters=self._parameters if parameters is UNSET else parameters,
        )

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy(
            index=copy.deepcopy(self._index, memo),
            content=copy.deepcopy(self._content, memo),
            parameters=copy.deepcopy(self._parameters, memo),
        )

    @classmethod
    def simplified(cls, index, content, *, parameters=None):
        is_cat = parameters is not None and parameters.get("__array__") == "categorical"

        if content.is_union and not is_cat:
            return content._union_of_optionarrays(index, parameters)

        elif content.is_indexed or content.is_option:
            backend = content.backend
            if content.is_indexed:
                inner = content.index
            else:
                inner = content.to_IndexedOptionArray64().index
            result = ak.index.Index64.empty(index.length, nplike=backend.nplike)

            backend.maybe_kernel_error(
                backend[
                    "awkward_IndexedArray_simplify",
                    result.dtype.type,
                    index.dtype.type,
                    inner.dtype.type,
                ](
                    result.data,
                    index.data,
                    index.length,
                    inner.data,
                    inner.length,
                )
            )
            return IndexedOptionArray(
                result,
                content.content,
                parameters=parameters_union(content._parameters, parameters),
            )

        else:
            return cls(index, content, parameters=parameters)

    def _form_with_key(
        self, getkey: Callable[[Content], str | None]
    ) -> IndexedOptionForm:
        form_key = getkey(self)
        return self.form_cls(
            self._index.form,
            self._content._form_with_key(getkey),
            parameters=self._parameters,
            form_key=form_key,
        )

    def _form_with_key_path(self, path: FormKeyPathT) -> IndexedOptionForm:
        return self.form_cls(
            self._index.form,
            self._content._form_with_key_path((*path, None)),
            parameters=self._parameters,
            form_key=repr(path),
        )

    def _to_buffers(
        self,
        form: Form,
        getkey: Callable[[Content, Form, str], str],
        container: MutableMapping[str, ArrayLike],
        backend: Backend,
        byteorder: str,
    ):
        assert isinstance(form, self.form_cls)
        key = getkey(self, form, "index")
        container[key] = ak._util.native_to_byteorder(
            self._index.raw(backend.nplike), byteorder
        )
        self._content._to_buffers(form.content, getkey, container, backend, byteorder)

    def _to_typetracer(self, forget_length: bool) -> Self:
        index = self._index.to_nplike(TypeTracer.instance())
        return IndexedOptionArray(
            index.forget_length() if forget_length else index,
            self._content._to_typetracer(forget_length),
            parameters=self._parameters,
        )

    def _touch_data(self, recursive: bool):
        self._index._touch_data()
        if recursive:
            self._content._touch_data(recursive)

    def _touch_shape(self, recursive: bool):
        self._index._touch_shape()
        if recursive:
            self._content._touch_shape(recursive)

    @property
    def length(self) -> ShapeItem:
        return self._index.length

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<IndexedOptionArray len="]
        out.append(repr(str(ak._util.maybe_length_of(self))))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._index._repr(indent + "    ", "<index>", "</index>\n"))
        out.append(self._content._repr(indent + "    ", "<content>", "</content>\n"))
        out.append(indent + "</IndexedOptionArray>")
        out.append(post)
        return "".join(out)

    def to_IndexedOptionArray64(self) -> IndexedOptionArray:
        if self._index.dtype == np.dtype(np.int64):
            return self
        else:
            return IndexedOptionArray(
                self._index.to64(),
                self._content,
                parameters=self._parameters,
            )

    def to_ByteMaskedArray(self, valid_when):
        mask = ak.index.Index8(self.mask_as_bool(valid_when))

        carry = self._index.data
        too_negative = carry < -1
        if self._backend.nplike.known_data and self._backend.nplike.any(too_negative):
            carry = carry.copy()
            carry[too_negative] = -1
        carry = ak.index.Index(carry)

        if self._content.length is not unknown_length and self._content.length == 0:
            content = self._content.form.length_one_array(backend=self._backend)._carry(
                carry, False
            )
        else:
            content = self._content._carry(carry, False)

        return ak.contents.ByteMaskedArray(
            mask, content, valid_when, parameters=self._parameters
        )

    def to_BitMaskedArray(self, valid_when, lsb_order):
        return self.to_ByteMaskedArray(valid_when).to_BitMaskedArray(
            valid_when, lsb_order
        )

    def mask_as_bool(self, valid_when: bool = True) -> ArrayLike:
        if valid_when:
            return self._index.raw(self._backend.nplike) >= 0
        else:
            return self._index.raw(self._backend.nplike) < 0

    def _getitem_nothing(self):
        return self._content._getitem_range(0, 0)

    def _is_getitem_at_placeholder(self) -> bool:
        if isinstance(self._index.data, PlaceholderArray):
            return True
        return self._content._is_getitem_at_placeholder()

    def _is_getitem_at_virtual(self) -> bool:
        is_virtual = (
            isinstance(self._index.data, VirtualNDArray)
            and not self._index.data.is_materialized
        )
        if is_virtual:
            return True
        return self._content._is_getitem_at_virtual()

    def _getitem_at(self, where: IndexType):
        if not self._backend.nplike.known_data:
            self._touch_data(recursive=False)
            return MaybeNone(self._content._getitem_at(where))

        if where < 0:
            where += self.length
        if self._backend.nplike.known_data and not 0 <= where < self.length:
            raise ak._errors.index_error(self, where)
        if self._index[where] < 0:
            return None
        else:
            return self._content._getitem_at(self._index[where])

    def _getitem_range(self, start: IndexType, stop: IndexType) -> Content:
        if not self._backend.nplike.known_data:
            self._touch_shape(recursive=False)
            return self

        # in non-typetracer mode (and if all lengths are known) we can check if the slice is a no-op
        # (i.e. slicing the full array) and shortcut to avoid noticeable python overhead
        if self._backend.nplike.known_data and (start == 0 and stop == self.length):
            return self

        return IndexedOptionArray(
            self._index[start:stop], self._content, parameters=self._parameters
        )

    def _getitem_field(
        self, where: str | SupportsIndex, only_fields: tuple[str, ...] = ()
    ) -> Content:
        return IndexedOptionArray.simplified(
            self._index,
            self._content._getitem_field(where, only_fields),
            parameters=None,
        )

    def _getitem_fields(
        self, where: list[str | SupportsIndex], only_fields: tuple[str, ...] = ()
    ) -> Content:
        return IndexedOptionArray.simplified(
            self._index,
            self._content._getitem_fields(where, only_fields),
            parameters=None,
        )

    def _carry(self, carry: Index, allow_lazy: bool) -> IndexedOptionArray:
        assert isinstance(carry, ak.index.Index)

        try:
            nextindex = self._index[carry.data]
        except IndexError as err:
            raise ak._errors.index_error(self, carry.data, str(err)) from err

        return IndexedOptionArray(nextindex, self._content, parameters=self._parameters)

    def _nextcarry_outindex(self):
        backend = self._backend
        nplike = backend.nplike

        _numnull = ak.index.Index64.empty(1, nplike=backend.nplike)
        assert _numnull.nplike is nplike and self._index.nplike is nplike
        self._backend.maybe_kernel_error(
            backend[
                "awkward_IndexedArray_numnull",
                _numnull.dtype.type,
                self._index.dtype.type,
            ](
                _numnull.data,
                self._index.data,
                self._index.length,
            )
        )
        numnull = nplike.index_as_shape_item(_numnull[0])
        nextcarry = ak.index.Index64.empty(
            self._index.length - numnull,
            nplike,
        )
        outindex = ak.index.Index.empty(
            self._index.length,
            nplike,
            dtype=self._index.dtype,
        )

        assert (
            nextcarry.nplike is nplike
            and outindex.nplike is nplike
            and self._index.nplike is nplike
        )
        self._backend.maybe_kernel_error(
            backend[
                "awkward_IndexedArray_getitem_nextcarry_outindex",
                nextcarry.dtype.type,
                outindex.dtype.type,
                self._index.dtype.type,
            ](
                nextcarry.data,
                outindex.data,
                self._index.data,
                self._index.length,
                self._content.length,
            )
        )

        return numnull, nextcarry, outindex

    def _getitem_next_jagged_generic(self, slicestarts, slicestops, slicecontent, tail):
        slicestarts = slicestarts.to_nplike(self._backend.nplike)
        slicestops = slicestops.to_nplike(self._backend.nplike)

        if self._backend.nplike.known_data and slicestarts.length != self.length:
            raise ak._errors.index_error(
                self,
                ak.contents.ListArray(
                    slicestarts, slicestops, slicecontent, parameters=None
                ),
                f"cannot fit jagged slice with length {slicestarts.length} into {type(self).__name__} of size {self.length}",
            )

        numnull, nextcarry, outindex = self._nextcarry_outindex()

        reducedstarts = ak.index.Index64.empty(
            self.length - numnull,
            nplike=self._backend.nplike,
        )
        reducedstops = ak.index.Index64.empty(
            self.length - numnull,
            nplike=self._backend.nplike,
        )
        assert (
            outindex.nplike is self._backend.nplike
            and slicestarts.nplike is self._backend.nplike
            and slicestops.nplike is self._backend.nplike
            and reducedstarts.nplike is self._backend.nplike
            and reducedstops.nplike is self._backend.nplike
        )
        self._maybe_index_error(
            self._backend[
                "awkward_MaskedArray_getitem_next_jagged_project",
                outindex.dtype.type,
                slicestarts.dtype.type,
                slicestops.dtype.type,
                reducedstarts.dtype.type,
                reducedstops.dtype.type,
            ](
                outindex.data,
                slicestarts.data,
                slicestops.data,
                reducedstarts.data,
                reducedstops.data,
                self.length,
            ),
            slicer=ak.contents.ListArray(slicestarts, slicestops, slicecontent),
        )
        next = self._content._carry(nextcarry, True)
        out = next._getitem_next_jagged(reducedstarts, reducedstops, slicecontent, tail)

        return ak.contents.IndexedOptionArray.simplified(
            outindex, out, parameters=self._parameters
        )

    def _getitem_next_jagged(
        self, slicestarts: Index, slicestops: Index, slicecontent: Content, tail
    ) -> Content:
        return self._getitem_next_jagged_generic(
            slicestarts, slicestops, slicecontent, tail
        )

    def _getitem_next(
        self,
        head: SliceItem | tuple,
        tail: tuple[SliceItem, ...],
        advanced: Index | None,
    ) -> Content:
        if head is NO_HEAD:
            return self

        elif is_integer_like(head) or isinstance(
            head, (slice, ak.index.Index64, ak.contents.ListOffsetArray)
        ):
            _nexthead, _nexttail = ak._slicing.head_tail(tail)

            _numnull, nextcarry, outindex = self._nextcarry_outindex()

            next = self._content._carry(nextcarry, True)
            out = next._getitem_next(head, tail, advanced)
            return IndexedOptionArray.simplified(
                outindex, out, parameters=self._parameters
            )

        elif isinstance(head, str):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list) and isinstance(head[0], str):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def project(self, mask=None):
        if mask is not None:
            if self._backend.nplike.known_data and self._index.length != mask.length:
                raise ValueError(
                    f"mask length ({mask.length()}) is not equal to {type(self).__name__} length ({self._index.length})"
                )
            nextindex = ak.index.Index64.empty(self._index.length, self._backend.nplike)
            assert (
                nextindex.nplike is self._backend.nplike
                and mask.nplike is self._backend.nplike
                and self._index.nplike is self._backend.nplike
            )
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_IndexedArray_overlay_mask",
                    nextindex.dtype.type,
                    mask.dtype.type,
                    self._index.dtype.type,
                ](
                    nextindex.data,
                    mask.data,
                    self._index.data,
                    self._index.length,
                )
            )
            next = ak.contents.IndexedOptionArray(
                nextindex, self._content, parameters=self._parameters
            )
            return next.project()
        else:
            _numnull = ak.index.Index64.empty(1, self._backend.nplike)
            assert (
                _numnull.nplike is self._backend.nplike
                and self._index.nplike is self._backend.nplike
            )
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_IndexedArray_numnull",
                    _numnull.dtype.type,
                    self._index.dtype.type,
                ](
                    _numnull.data,
                    self._index.data,
                    self._index.length,
                )
            )
            numnull = self._backend.nplike.index_as_shape_item(_numnull[0])
            nextcarry = ak.index.Index64.empty(
                self.length - numnull,
                self._backend.nplike,
            )

            assert (
                nextcarry.nplike is self._backend.nplike
                and self._index.nplike is self._backend.nplike
            )
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_IndexedArray_flatten_nextcarry",
                    nextcarry.dtype.type,
                    self._index.dtype.type,
                ](
                    nextcarry.data,
                    self._index.data,
                    self._index.length,
                    self._content.length,
                )
            )

            return self._content._carry(nextcarry, True)

    def _offsets_and_flattened(self, axis: int, depth: int) -> tuple[Index, Content]:
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            raise AxisError("axis=0 not allowed for flatten")
        else:
            numnull, nextcarry, outindex = self._nextcarry_outindex()
            next = self._content._carry(nextcarry, False)

            offsets, flattened = next._offsets_and_flattened(axis, depth)

            if offsets.length is not unknown_length and offsets.length == 0:
                return (
                    offsets,
                    ak.contents.IndexedOptionArray(
                        outindex, flattened, parameters=self._parameters
                    ),
                )

            else:
                outoffsets = ak.index.Index64.empty(
                    offsets.length + numnull,
                    self._backend.nplike,
                    dtype=np.int64,
                )

                assert (
                    outoffsets.nplike is self._backend.nplike
                    and outindex.nplike is self._backend.nplike
                    and offsets.nplike is self._backend.nplike
                )
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_IndexedArray_flatten_none2empty",
                        outoffsets.dtype.type,
                        outindex.dtype.type,
                        offsets.dtype.type,
                    ](
                        outoffsets.data,
                        outindex.data,
                        outindex.length,
                        offsets.data,
                        offsets.length,
                    )
                )
                return (outoffsets, flattened)

    def _mergeable_next(
        self,
        other: Content,
        mergebool: bool,
        mergecastable: Literal["same_kind", "equiv", "family"],
    ) -> bool:
        # Is the other content is an identity, or a union?
        if other.is_identity_like or other.is_union:
            return True
        # Is the other array indexed or optional?
        elif other.is_option or other.is_indexed:
            return self._content._mergeable_next(
                other.content, mergebool, mergecastable
            )
        else:
            return self._content._mergeable_next(other, mergebool, mergecastable)

    def _merging_strategy(self, others):
        if len(others) == 0:
            raise ValueError(
                "to merge this array with 'others', at least one other must be provided"
            )

        head = [self]
        tail = []

        it_others = iter(others)
        for other in it_others:
            if isinstance(other, ak.contents.UnionArray):
                tail.append(other)
                tail.extend(it_others)
                break
            else:
                head.append(other)

        if any(x.backend.nplike.known_data for x in head + tail) and not all(
            x.backend.nplike.known_data for x in head + tail
        ):
            raise RuntimeError

        return head, tail

    def _reverse_merge(self, other):
        if isinstance(other, ak.contents.EmptyArray):
            return self

        # FIXME: support categorical-categorical merging
        if (
            other.is_indexed
            and other.parameter("__array__")
            == self.parameter("__array__")
            == "categorical"
        ):
            raise NotImplementedError(
                "merging categorical arrays is currently not implemented. "
                "Use `ak.enforce_type` to drop the categorical type and use general merging."
            )

        theirlength = other.length
        mylength = self.length
        index = ak.index.Index64.empty(
            theirlength + mylength,
            self._backend.nplike,
        )

        content = other._mergemany([self._content])

        # Fill index::0→theirlength with arange(theirlength)
        assert index.nplike is self._backend.nplike
        self._backend.maybe_kernel_error(
            self._backend["awkward_IndexedArray_fill_count", index.dtype.type](
                index.data,
                0,
                theirlength,
                0,
            )
        )

        # Fill index::theirlength->end with self.index[:mylength]+theirlength
        assert (
            index.nplike is self._backend.nplike
            and self.index.nplike is self._backend.nplike
        )
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_IndexedArray_fill",
                index.dtype.type,
                self.index.dtype.type,
            ](
                index.data,
                theirlength,
                self.index.data,
                mylength,
                theirlength,
            )
        )
        # We can directly merge with other options and indexed types, but we must merge parameters
        if other.is_option or other.is_indexed:
            parameters = parameters_union(self._parameters, other._parameters)
        # Otherwise, this option parameters win out
        else:
            parameters = self._parameters

        return ak.contents.IndexedOptionArray.simplified(
            index, content, parameters=parameters
        )

    def _mergemany(self, others: Sequence[Content]) -> Content:
        if len(others) == 0:
            return self

        head, tail = self._merging_strategy(others)

        total_length = 0
        for array in head:
            total_length += array.length

        contents = []
        contentlength_so_far = 0
        length_so_far = 0
        nextindex = ak.index.Index64.empty(total_length, self._backend.nplike)
        parameters = self._parameters

        for array in head:
            if isinstance(array, ak.contents.EmptyArray):
                continue

            if isinstance(
                array,
                (
                    ak.contents.ByteMaskedArray,
                    ak.contents.BitMaskedArray,
                    ak.contents.UnmaskedArray,
                ),
            ):
                array = array.to_IndexedOptionArray64()

            if isinstance(
                array, (ak.contents.IndexedOptionArray, ak.contents.IndexedArray)
            ):
                array = array._trim()  # see: #3185 and #3119
                # If we're merging an option, then merge parameters before pulling out `content`
                parameters = parameters_intersect(parameters, array._parameters)
                contents.append(array.content)
                array_index = array.index
                assert (
                    nextindex.nplike is self._backend.nplike
                    and array_index.nplike is self._backend.nplike
                )
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_IndexedArray_fill",
                        nextindex.dtype.type,
                        array_index.dtype.type,
                    ](
                        nextindex.data,
                        length_so_far,
                        array_index.data,
                        array.length,
                        contentlength_so_far,
                    )
                )
                contentlength_so_far += array.content.length

                length_so_far += array.length

            else:
                contents.append(array)
                assert nextindex.nplike is self._backend.nplike
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_IndexedArray_fill_count",
                        nextindex.dtype.type,
                    ](
                        nextindex.data,
                        length_so_far,
                        array.length,
                        contentlength_so_far,
                    )
                )
                contentlength_so_far += array.length

                length_so_far += array.length

                # Categoricals may only survive if all contents are categorical
                if (
                    parameters is not None
                    and parameters.get("__array__") == "categorical"
                ):
                    parameters = {**parameters}
                    del parameters["__array__"]

        tail_contents = contents[1:]
        nextcontent = contents[0]._mergemany(tail_contents)
        next = ak.contents.IndexedOptionArray(
            nextindex, nextcontent, parameters=parameters
        )

        # FIXME: support categorical merging?
        if parameters is not None and parameters.get("__array__") == "categorical":
            raise NotImplementedError(
                "merging categorical arrays is currently not implemented. "
                "Use `ak.enforce_type` to drop the categorical type and use general merging."
            )

        if len(tail) == 0:
            return next

        reversed = tail[0]._reverse_merge(next)
        if len(tail) == 1:
            return reversed
        else:
            return reversed._mergemany(tail[1:])

    def _fill_none(self, value: Content) -> Content:
        if value.backend.nplike.known_data and value.length != 1:
            raise ValueError(
                f"fill_none value length ({value.length}) is not equal to 1"
            )

        contents = [self._content, value]
        tags = ak.index.Index8(self.mask_as_bool(valid_when=False))
        index = ak.index.Index64.empty(tags.length, self._backend.nplike)

        assert (
            index.nplike is self._backend.nplike
            and self._index.nplike is self._backend.nplike
        )
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_UnionArray_fillna", index.dtype.type, self._index.dtype.type
            ](index.data, self._index.data, tags.length)
        )
        return ak.contents.UnionArray.simplified(
            tags,
            index,
            contents,
            parameters=self._parameters,
            mergebool=True,
        )

    def _local_index(self, axis, depth):
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            return self._local_index_axis0()
        else:
            _, nextcarry, outindex = self._nextcarry_outindex()

            next = self._content._carry(nextcarry, False)
            out = next._local_index(axis, depth)
            out2 = ak.contents.IndexedOptionArray(
                outindex, out, parameters=self._parameters
            )
            return out2

    def _is_subrange_equal(self, starts, stops, length, sorted=True):
        nextstarts = ak.index.Index64.empty(length, self._backend.nplike)
        nextstops = ak.index.Index64.empty(length, self._backend.nplike)

        subranges_length = ak.index.Index64.empty(1, self._backend.nplike)
        assert (
            self._index.nplike is self._backend.nplike
            and starts.nplike is self._backend.nplike
            and stops.nplike is self._backend.nplike
            and nextstarts.nplike is self._backend.nplike
            and nextstops.nplike is self._backend.nplike
            and subranges_length.nplike is self._backend.nplike
        )
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_IndexedArray_ranges_next_64",
                self._index.dtype.type,
                starts.dtype.type,
                stops.dtype.type,
                nextstarts.dtype.type,
                nextstops.dtype.type,
                subranges_length.dtype.type,
            ](
                self._index.data,
                starts.data,
                stops.data,
                length,
                nextstarts.data,
                nextstops.data,
                subranges_length.data,
            )
        )

        nextcarry = ak.index.Index64.empty(subranges_length[0], self._backend.nplike)
        assert (
            self._index.nplike is self._backend.nplike
            and starts.nplike is self._backend.nplike
            and stops.nplike is self._backend.nplike
            and nextcarry.nplike is self._backend.nplike
        )
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_IndexedArray_ranges_carry_next_64",
                self._index.dtype.type,
                starts.dtype.type,
                stops.dtype.type,
                nextcarry.dtype.type,
            ](
                self._index.data,
                starts.data,
                stops.data,
                length,
                nextcarry.data,
            )
        )

        next = self._content._carry(nextcarry, False)
        if nextstarts.length is not unknown_length and nextstarts.length > 1:
            return next._is_subrange_equal(nextstarts, nextstops, nextstarts.length)
        else:
            return next._subranges_equal(
                nextstarts, nextstops, nextstarts.length, False
            )

    def _numbers_to_type(self, name, including_unknown):
        return ak.contents.IndexedOptionArray(
            self._index,
            self._content._numbers_to_type(name, including_unknown),
            parameters=self._parameters,
        )

    def _is_unique(self, negaxis, starts, parents, outlength):
        if self._index.length is not unknown_length and self._index.length == 0:
            return True

        projected = self.project()
        return projected._is_unique(negaxis, starts, parents, outlength)

    def _unique(self, negaxis, starts, parents, outlength):
        branch, depth = self.branch_depth

        inject_nones = (
            True if not branch and (negaxis is not None and negaxis != depth) else False
        )

        index_length = self._index.length

        next, nextparents, _numnull, _outindex = self._rearrange_prepare_next(parents)

        out = next._unique(
            negaxis,
            starts,
            nextparents,
            outlength,
        )

        if branch or (negaxis is not None and negaxis != depth):
            nextoutindex = ak.index.Index64.empty(parents.length, self._backend.nplike)
            assert (
                nextoutindex.nplike is self._backend.nplike
                and starts.nplike is self._backend.nplike
                and parents.nplike is self._backend.nplike
                and nextparents.nplike is self._backend.nplike
            )

            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_IndexedArray_local_preparenext_64",
                    nextoutindex.dtype.type,
                    starts.dtype.type,
                    parents.dtype.type,
                    nextparents.dtype.type,
                ](
                    nextoutindex.data,
                    starts.data,
                    parents.data,
                    parents.length,
                    nextparents.data,
                    nextparents.length,
                )
            )

            return ak.contents.IndexedOptionArray.simplified(
                nextoutindex, out, parameters=self._parameters
            )

        if isinstance(out, ak.contents.ListOffsetArray):
            newnulls = ak.index.Index64.empty(self._index.length, self._backend.nplike)
            _len_newnulls = ak.index.Index64.empty(1, self._backend.nplike)
            assert (
                newnulls.nplike is self._backend.nplike
                and _len_newnulls.nplike is self._backend.nplike
                and self._index.nplike is self._backend.nplike
            )
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_IndexedArray_numnull_parents",
                    newnulls.dtype.type,
                    _len_newnulls.dtype.type,
                    self._index.dtype.type,
                ](
                    newnulls.data,
                    _len_newnulls.data,
                    self._index.data,
                    index_length,
                )
            )
            len_newnulls = self._backend.nplike.index_as_shape_item(_len_newnulls[0])

            newindex = ak.index.Index64.empty(
                self._backend.nplike.index_as_shape_item(out._offsets[-1])
                + len_newnulls,
                self._backend.nplike,
            )
            newoffsets = ak.index.Index64.empty(
                out._offsets.length, self._backend.nplike
            )
            assert (
                newindex.nplike is self._backend.nplike
                and newoffsets.nplike is self._backend.nplike
                and out._offsets.nplike is self._backend.nplike
                and newnulls.nplike is self._backend.nplike
            )
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_IndexedArray_unique_next_index_and_offsets_64",
                    newindex.dtype.type,
                    newoffsets.dtype.type,
                    out._offsets.dtype.type,
                    newnulls.dtype.type,
                ](
                    newindex.data,
                    newoffsets.data,
                    out._offsets.data,
                    newnulls.data,
                    starts.length,
                )
            )

            out = ak.contents.IndexedOptionArray.simplified(
                newindex[: newoffsets[-1]], out._content, parameters=self._parameters
            )
            return ak.contents.ListOffsetArray(
                newoffsets, out, parameters=self._parameters
            )

        if isinstance(out, ak.contents.NumpyArray):
            # We do not pass None values to the content, so it will have computed
            # the unique _non null_ values. We therefore need to account for None
            # values in the result. We do this by creating an IndexedOptionArray
            # and tacking the index -1 onto the end
            nextoutindex = ak.index.Index64.empty(
                out.length + 1,
                self._backend.nplike,
            )
            assert nextoutindex.nplike is self._backend.nplike
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_IndexedArray_numnull_unique_64",
                    nextoutindex.dtype.type,
                ](
                    nextoutindex.data,
                    out.length,
                )
            )

            return ak.contents.IndexedOptionArray.simplified(
                nextoutindex, out, parameters=self._parameters
            )

        if inject_nones:
            out = ak.contents.RegularArray(
                out, out.length, 0, parameters=self._parameters
            )

        return out

    def _rearrange_nextshifts(self, nextparents, shifts):
        nextshifts = ak.index.Index64.empty(nextparents.length, self._backend.nplike)
        assert nextshifts.nplike is self._backend.nplike

        if shifts is None:
            assert (
                nextshifts.nplike is self._backend.nplike
                and self._index.nplike is self._backend.nplike
            )
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_IndexedArray_reduce_next_nonlocal_nextshifts_64",
                    nextshifts.dtype.type,
                    self._index.dtype.type,
                ](
                    nextshifts.data,
                    self._index.data,
                    self._index.length,
                )
            )
        else:
            assert (
                nextshifts.nplike is self._backend.nplike
                and self._index.nplike is self._backend.nplike
                and shifts.nplike is self._backend.nplike
            )
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_IndexedArray_reduce_next_nonlocal_nextshifts_fromshifts_64",
                    nextshifts.dtype.type,
                    self._index.dtype.type,
                    shifts.dtype.type,
                ](
                    nextshifts.data,
                    self._index.data,
                    self._index.length,
                    shifts.data,
                )
            )
        return nextshifts

    def _rearrange_prepare_next(self, parents):
        assert (
            self._index.nplike is self._backend.nplike
            and parents.nplike is self._backend.nplike
        )
        index_length = self._index.length
        _numnull = ak.index.Index64.empty(1, self._backend.nplike)
        assert _numnull.nplike is self._backend.nplike

        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_IndexedArray_numnull",
                _numnull.dtype.type,
                self._index.dtype.type,
            ](
                _numnull.data,
                self._index.data,
                index_length,
            )
        )
        numnull = self._backend.nplike.index_as_shape_item(_numnull[0])

        next_length = index_length - numnull
        nextparents = ak.index.Index64.empty(next_length, self._backend.nplike)
        nextcarry = ak.index.Index64.empty(next_length, self._backend.nplike)
        outindex = ak.index.Index64.empty(index_length, self._backend.nplike)
        assert (
            nextcarry.nplike is self._backend.nplike
            and nextparents.nplike is self._backend.nplike
            and outindex.nplike is self._backend.nplike
        )
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_IndexedArray_reduce_next_64",
                nextcarry.dtype.type,
                nextparents.dtype.type,
                outindex.dtype.type,
                self._index.dtype.type,
                parents.dtype.type,
            ](
                nextcarry.data,
                nextparents.data,
                outindex.data,
                self._index.data,
                parents.data,
                index_length,
            )
        )
        next = self._content._carry(nextcarry, False)
        return next, nextparents, numnull, outindex

    def _argsort_next(
        self, negaxis, starts, shifts, parents, outlength, ascending, stable
    ):
        assert (
            starts.nplike is self._backend.nplike
            and parents.nplike is self._backend.nplike
            and self._index.nplike is self._backend.nplike
        )

        branch, depth = self.branch_depth

        next, nextparents, numnull, outindex = self._rearrange_prepare_next(parents)

        if (not branch) and negaxis == depth:
            nextshifts = self._rearrange_nextshifts(nextparents, shifts)
        else:
            nextshifts = None

        out = next._argsort_next(
            negaxis, starts, nextshifts, nextparents, outlength, ascending, stable
        )

        # `next._argsort_next` is given the non-None values. We choose to
        # sort None values to the end of the list, meaning we need to grow `out`
        # to account for these None values. First, we locate these nones within
        # their sublists
        nulls_merged = False
        nulls_index = ak.index.Index64.empty(numnull, self._backend.nplike)
        assert nulls_index.nplike is self._backend.nplike
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_IndexedArray_index_of_nulls",
                nulls_index.dtype.type,
                self._index.dtype.type,
                parents.dtype.type,
                starts.dtype.type,
            ](
                nulls_index.data,
                self._index.data,
                self._index.length,
                parents.data,
                starts.data,
            )
        )
        # If we wrap a NumpyArray (i.e., axis=-1), then we want `argmax` to return
        # the indices of each `None` value, rather than `None` itself.
        # We can test for this condition by seeing whether the NumpyArray of indices
        # is mergeable with our content (`out = next._argsort_next result`).
        # If so, try to concatenate them at the end of `out`.`
        nulls_index_content = ak.contents.NumpyArray(
            nulls_index.data, parameters=None, backend=self._backend
        )
        if out._mergeable_next(nulls_index_content, True, "same_kind"):
            out = out._mergemany([nulls_index_content])
            nulls_merged = True

        nextoutindex = ak.index.Index64.empty(parents.length, self._backend.nplike)
        assert (
            nextoutindex.nplike is self._backend.nplike
            and starts.nplike is self._backend.nplike
            and parents.nplike is self._backend.nplike
            and nextparents.nplike is self._backend.nplike
        )

        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_IndexedArray_local_preparenext_64",
                nextoutindex.dtype.type,
                starts.dtype.type,
                parents.dtype.type,
                nextparents.dtype.type,
            ](
                nextoutindex.data,
                starts.data,
                parents.data,
                parents.length,
                nextparents.data,
                nextparents.length,
            )
        )

        if nulls_merged:
            # awkward_IndexedArray_local_preparenext uses -1 to
            # indicate `None` values. Given that this code-path runs
            # only when the `None` value indices are explicitly stored in out,
            # we need to mapping the -1 values to their corresponding indices
            # in `out`
            assert nextoutindex.nplike is self._backend.nplike
            self._backend.maybe_kernel_error(
                self._backend["awkward_Index_nones_as_index", nextoutindex.dtype.type](
                    nextoutindex.data,
                    nextoutindex.length,
                )
            )
            # Drop the option type: we have ensured that we don't have any
            # -1 values in `nextoutindex` now!
            out = out._carry(nextoutindex, False).copy(
                parameters=parameters_union(out._parameters, self._parameters)
            )
        else:
            out = ak.contents.IndexedOptionArray.simplified(
                nextoutindex, out, parameters=self._parameters
            )

        inject_nones = (
            True
            if (
                (numnull is not unknown_length and numnull > 0)
                and not branch
                and negaxis != depth
            )
            else False
        )

        # If we want the None's at this depth to be injected
        # into the dense ([x y z None None]) rearranger result.
        # Here, we index the dense content with an index
        # that maps the values to the correct locations
        if inject_nones:
            return ak.contents.IndexedOptionArray.simplified(
                outindex, out, parameters=self._parameters
            )
        # Otherwise, if we are rearranging (e.g sorting) the contents of this layout,
        # then we do NOT want to return an optional layout,
        # OR we are branching
        else:
            return out

    def _sort_next(self, negaxis, starts, parents, outlength, ascending, stable):
        assert (
            starts.nplike is self._backend.nplike
            and parents.nplike is self._backend.nplike
        )
        branch, depth = self.branch_depth

        next, nextparents, _numnull, outindex = self._rearrange_prepare_next(parents)

        out = next._sort_next(
            negaxis, starts, nextparents, outlength, ascending, stable
        )

        nextoutindex = ak.index.Index64.empty(parents.length, self._backend.nplike)
        assert nextoutindex.nplike is self._backend.nplike

        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_IndexedArray_local_preparenext_64",
                nextoutindex.dtype.type,
                starts.dtype.type,
                parents.dtype.type,
                nextparents.dtype.type,
            ](
                nextoutindex.data,
                starts.data,
                parents.data,
                parents.length,
                nextparents.data,
                nextparents.length,
            )
        )
        out = ak.contents.IndexedOptionArray.simplified(
            nextoutindex, out, parameters=self._parameters
        )

        inject_nones = True if not branch and negaxis != depth else False

        # If we want the None's at this depth to be injected
        # into the dense ([x y z None None]) rearranger result.
        # Here, we index the dense content with an index
        # that maps the values to the correct locations
        if inject_nones:
            return ak.contents.IndexedOptionArray.simplified(
                outindex, out, parameters=self._parameters
            )
        # Otherwise, if we are rearranging (e.g sorting) the contents of this layout,
        # then we do NOT want to return an optional layout
        # OR we are branching
        else:
            return out

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
        behavior,
    ):
        branch, depth = self.branch_depth

        next, nextparents, _numnull, outindex = self._rearrange_prepare_next(parents)

        if reducer.needs_position and (not branch and negaxis == depth):
            nextshifts = self._rearrange_nextshifts(nextparents, shifts)
        else:
            nextshifts = None

        out = next._reduce_next(
            reducer,
            negaxis,
            starts,
            nextshifts,
            nextparents,
            outlength,
            mask,
            keepdims,
            behavior,
        )

        # If we are reducing the contents of this layout,
        # then we do NOT want to return an optional layout
        if not branch and negaxis == depth:
            return out
        else:
            if isinstance(out, ak.contents.RegularArray):
                out_content = out.content
            elif isinstance(out, ak.contents.ListOffsetArray):
                # The `outindex` that will index into `out_content` is 0-based, so we should ensure that we normalise
                # the list content to start at the first offset.
                out_content = out.content[out.offsets[0] :]
            else:
                raise AssertionError(
                    "reduce_next with unbranching depth > negaxis is only "
                    "expected to return RegularArray or ListOffsetArray or "
                    "IndexedOptionArray; "
                    f"instead, it returned {type(out).__name__}"
                )

            if starts.nplike.known_data and starts.length > 0 and starts[0] != 0:
                raise AssertionError(
                    "reduce_next with unbranching depth > negaxis expects a "
                    f"ListOffsetArray whose offsets start at zero ({starts[0]})"
                )
            # In this branch, we're above the axis at which the reduction takes place.
            # `next._reduce_next` is therefore expected to return a list/regular layout
            # node. As detailed in `RegularArray._reduce_next`, `_reduce_next` wraps the
            # reduction in a list-type of length `outlength` before returning to the caller,
            # which effectively means that the reduction of *this* layout corresponds to the
            # child of the returned `next._reduce_next(...)`, i.e. `out.content`. So, we unpack
            # the returned list type and wrap its child by a new `IndexedOptionArray`, before
            # re-wrapping the result to have the length and starts requested by the caller.
            if starts.length is unknown_length:
                outoffsets = ak.index.Index64.empty(
                    unknown_length, self._backend.nplike
                )
            else:
                outoffsets = ak.index.Index64.empty(
                    starts.length + 1,
                    self._backend.nplike,
                )
            assert (
                outoffsets.nplike is self._backend.nplike
                and starts.nplike is self._backend.nplike
            )
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_IndexedArray_reduce_next_fix_offsets_64",
                    outoffsets.dtype.type,
                    starts.dtype.type,
                ](
                    outoffsets.data,
                    starts.data,
                    starts.length,
                    outindex.length,
                )
            )

            # Apply `outindex` to appropriate content
            inner = ak.contents.IndexedOptionArray.simplified(
                outindex, out_content, parameters=self._parameters
            )

            # Re-wrap content
            return ak.contents.ListOffsetArray(
                outoffsets, inner, parameters=self._parameters
            )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        else:
            _, nextcarry, outindex = self._nextcarry_outindex()
            next = self._content._carry(nextcarry, True)
            out = next._combinations(
                n, replacement, recordlookup, parameters, axis, depth
            )
            return IndexedOptionArray.simplified(outindex, out, parameters=parameters)

    def _validity_error(self, path):
        if self.parameter("__array__") == "categorical":
            if not ak._do.is_unique(self._content):
                return f'at {path} ("{type(self)}"): __array__ = "categorical" requires contents to be unique'

        assert self.index.nplike is self._backend.nplike
        error = self._backend["awkward_IndexedArray_validity", self.index.dtype.type](
            self.index.data, self.index.length, self._content.length, True
        )
        if error is not None and error.str is not None:
            if error.filename is None:
                filename = ""
            else:
                filename = " (in compiled code: " + error.filename.decode(
                    errors="surrogateescape"
                ).lstrip("\n").lstrip("(")
            message = error.str.decode(errors="surrogateescape")
            return f'at {path} ("{type(self)}"): {message} at i={error.id}{filename}'

        else:
            return self._content._validity_error(path + ".content")

    def _nbytes_part(self):
        return self.index._nbytes_part() + self.content._nbytes_part()

    def _pad_none(self, target, axis, depth, clip):
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            return self._pad_none_axis0(target, clip)
        elif posaxis is not None and posaxis + 1 == depth + 1:
            mask = ak.index.Index8(self.mask_as_bool(valid_when=False))
            index = ak.index.Index64.empty(mask.length, self._backend.nplike)
            assert (
                index.nplike is self._backend.nplike
                and mask.nplike is self._backend.nplike
            )
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_IndexedOptionArray_rpad_and_clip_mask_axis1",
                    index.dtype.type,
                    mask.dtype.type,
                ](index.data, mask.data, mask.length)
            )
            next = self.project()._pad_none(target, axis, depth, clip)
            return ak.contents.IndexedOptionArray.simplified(
                index, next, parameters=self._parameters
            )
        else:
            return ak.contents.IndexedOptionArray(
                self._index,
                self._content._pad_none(target, axis, depth, clip),
                parameters=self._parameters,
            )

    def _to_arrow(
        self,
        pyarrow: Any,
        mask_node: Content | None,
        validbytes: Content | None,
        length: int,
        options: ToArrowOptions,
    ):
        (index,) = maybe_materialize(numpy.asarray(self._index.data, copy=True))
        this_validbytes = self.mask_as_bool(valid_when=True)
        index[~this_validbytes] = 0

        if self.parameter("__array__") == "categorical":
            # The new IndexedArray will have this parameter, but the rest
            # will be in the AwkwardArrowType.mask_parameters.
            next_parameters = {"__array__": "categorical"}
        else:
            next_parameters = None

        next = ak.contents.IndexedArray(
            ak.index.Index(index), self._content, parameters=next_parameters
        )
        return next._to_arrow(
            pyarrow,
            self,
            ak._connect.pyarrow.and_validbytes(validbytes, this_validbytes),
            length,
            options,
        )

    def _to_cudf(self, cudf: Any, mask: Content | None, length: int):
        return self.to_ByteMaskedArray(True)._to_cudf(cudf, mask, length)

    def _to_backend_array(self, allow_missing, backend):
        nplike = backend.nplike
        nplike = backend.nplike

        content = self.project()._to_backend_array(allow_missing, backend)
        shape = (self.length, *content.shape[1:])
        data = nplike.empty(shape, dtype=content.dtype)
        mask0 = nplike.asarray(self.mask_as_bool(valid_when=False)).view(np.bool_)
        if nplike.any(mask0):
            if allow_missing:
                mask = nplike.broadcast_to(
                    nplike.reshape(mask0, (shape[0],) + (1,) * (len(shape) - 1)),
                    shape,
                )
                if isinstance(content, nplike.ma.MaskedArray):
                    mask1 = nplike.ma.getmaskarray(content)
                    mask = nplike.asarray(mask, copy=True)
                    mask[nplike.logical_not(mask0)] |= mask1

                data[nplike.logical_not(mask0)] = content

                if content.dtype.names is not None:
                    missing_data = tuple(
                        create_missing_data(each_dtype, backend)
                        for each_dtype, _ in content.dtype.fields.values()
                    )
                else:
                    missing_data = create_missing_data(content.dtype, backend)

                data[mask0] = missing_data

                return nplike.ma.MaskedArray(data, mask)
            else:
                raise ValueError(
                    "Content.to_nplike cannot convert 'None' values to "
                    "np.ma.MaskedArray unless the "
                    "'allow_missing' parameter is set to True"
                )
        else:
            if allow_missing:
                return nplike.ma.MaskedArray(content)
            else:
                return content

    def _remove_structure(
        self, backend: Backend, options: RemoveStructureOptions
    ) -> list[Content]:
        branch, depth = self.branch_depth
        if branch or options["drop_nones"] or depth > 1:
            return self.project()._remove_structure(backend, options)
        else:
            return [self]

    def _drop_none(self) -> Content:
        return self.project()

    def _recursively_apply(
        self,
        action: ImplementsApplyAction,
        depth: int,
        depth_context: Mapping[str, Any] | None,
        lateral_context: Mapping[str, Any] | None,
        options: ApplyActionOptions,
    ) -> Content | None:
        if (
            self._backend.nplike.known_data
            and self._backend.nplike.known_data
            and self._index.length != 0
        ):
            npindex = self._index.data
            npselect = npindex >= 0
            if self._backend.nplike.any(npselect):
                indexmin = self._backend.nplike.min(npindex[npselect])
                indexmax = self._backend.nplike.max(npindex)
                index = ak.index.Index(npindex - indexmin, nplike=self._backend.nplike)
                content = self._content[indexmin : indexmax + 1]
            else:
                index, content = self._index, self._content
        else:
            if not self._backend.nplike.known_data:
                self._touch_data(recursive=False)
            index, content = self._index, self._content

        if options["return_array"]:
            if options["return_simplified"]:
                make = IndexedOptionArray.simplified
            else:
                make = IndexedOptionArray

            def continuation():
                return make(
                    index,
                    content._recursively_apply(
                        action,
                        depth,
                        copy.copy(depth_context),
                        lateral_context,
                        options,
                    ),
                    parameters=self._parameters if options["keep_parameters"] else None,
                )

        else:

            def continuation():
                content._recursively_apply(
                    action,
                    depth,
                    copy.copy(depth_context),
                    lateral_context,
                    options,
                )

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            backend=self._backend,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def _to_packed(self, recursive: bool = True) -> Self:
        if self.parameter("__array__") == "categorical":
            content = self._content.to_packed(True) if recursive else self._content
            return ak.contents.IndexedOptionArray(
                self._index, content, parameters=self._parameters
            )

        else:
            nplike = self._backend.nplike
            original_index = self._index.data
            is_none = original_index < 0
            num_none = nplike.index_as_shape_item(nplike.count_nonzero(is_none))
            new_index = nplike.empty(self._index.length, dtype=self._index.dtype)
            if isinstance(nplike, Jax):
                new_index = new_index.at[is_none].set(-1)
                new_index = new_index.at[~is_none].set(
                    nplike.arange(
                        nplike.shape_item_as_index(new_index.size - num_none),
                        dtype=self._index.dtype,
                    )
                )
            else:
                new_index[is_none] = -1
                new_index[~is_none] = nplike.arange(
                    nplike.shape_item_as_index(new_index.size - num_none),
                    dtype=self._index.dtype,
                )
            projected = self.project()
            return ak.contents.IndexedOptionArray(
                ak.index.Index(new_index, nplike=self._backend.nplike),
                projected.to_packed(True) if recursive else projected,
                parameters=self._parameters,
            )

    def _to_list(self, behavior, json_conversions):
        if not self._backend.nplike.known_data:
            raise TypeError("cannot convert typetracer arrays to Python lists")

        out = self._to_list_custom(behavior, json_conversions)
        if out is not None:
            return out

        (index,) = maybe_materialize(self._index.raw(numpy))
        not_missing = index >= 0
        content = ak.to_backend(self._content, "cpu", highlevel=False)
        nextcontent = content._carry(ak.index.Index(index[not_missing]), False)
        out = nextcontent._to_list(behavior, json_conversions)

        for i, isvalid in enumerate(not_missing):
            if not isvalid:
                out.insert(i, None)

        return out

    def _to_backend(self, backend: Backend) -> Self:
        content = self._content.to_backend(backend)
        index = self._index.to_nplike(backend.nplike)
        return IndexedOptionArray(index, content, parameters=self._parameters)

    def _materialize(self, type_) -> Self:
        content = self._content.materialize(type_)
        index = self._index.materialize(type_)
        return IndexedOptionArray(index, content, parameters=self._parameters)

    @property
    def _is_all_materialized(self) -> bool:
        return self._content.is_all_materialized and self._index.is_all_materialized

    @property
    def _is_any_materialized(self) -> bool:
        return self._content.is_any_materialized or self._index.is_any_materialized

    def _is_equal_to(
        self, other: Self, index_dtype: bool, numpyarray: bool, all_parameters: bool
    ) -> bool:
        return (
            self._is_equal_to_generic(other, all_parameters)
            and self._index.is_equal_to(other.index, index_dtype, numpyarray)
            and self._content._is_equal_to(
                other.content, index_dtype, numpyarray, all_parameters
            )
        )

    def _trim(self) -> Self:
        nplike = self._backend.nplike

        if not nplike.known_data or self._index.length == 0:
            return self

        idx_buf = nplike.asarray(self._index.data, copy=True)
        only_positive = idx_buf >= 0

        # no positive index at all
        if not nplike.any(only_positive):
            return self

        min_idx = nplike.min(idx_buf[only_positive])
        max_idx = nplike.max(idx_buf[only_positive])
        idx_buf[only_positive] -= min_idx
        index = Index(idx_buf)

        # left and right trim
        content = self._content._getitem_range(min_idx, max_idx + 1)
        return IndexedOptionArray(index, content, parameters=self._parameters)


def create_missing_data(dtype, backend):
    """Create missing data based on the input dtype

    Missing data are represented differently based on the Numpy array
    dtype, this function returns the proper missing data representation
    given the input dtype
    """
    if np.issubdtype(dtype, np.bool_):
        return False
    elif np.issubdtype(dtype, np.floating):
        return np.nan
    elif np.issubdtype(dtype, np.complexfloating):
        return np.nan + np.nan * 1j
    elif np.issubdtype(dtype, np.integer):
        return np.iinfo(dtype).max
    elif np.issubdtype(dtype.type, np.datetime64) or np.issubdtype(
        dtype.type, np.timedelta64
    ):
        return backend.nplike.asarray([np.iinfo(np.int64).max], dtype=dtype)[0]
    elif np.issubdtype(dtype, np.str_):
        return ""
    elif np.issubdtype(dtype, np.bytes_):
        return b""
    else:
        raise AssertionError(f"unrecognized dtype: {dtype}")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

"""
Pyarrow extension classes: AwkwardArrowArray and AwkwardArrowType
See: https://arrow.apache.org/docs/python/extending_types.html
"""

from __future__ import annotations

import json

import pyarrow

from awkward._nplikes.numpy_like import NumpyMetadata

np = NumpyMetadata.instance()


class AwkwardArrowArray(pyarrow.ExtensionArray):
    def to_pylist(self, maps_as_pydicts=None):
        out = super().to_pylist()
        if (
            isinstance(self.type, AwkwardArrowType)
            and self.type.node_type == "RecordArray"
            and self.type.record_is_tuple is True
        ):
            for i, x in enumerate(out):
                if x is not None:
                    out[i] = tuple(x[str(j)] for j in range(len(x)))
        return out


class AwkwardArrowType(pyarrow.ExtensionType):
    def __init__(
        self,
        storage_type,
        mask_type,
        node_type,
        mask_parameters,
        node_parameters,
        record_is_tuple,
        record_is_scalar,
        is_nonnullable_nulltype=False,
        option_type=False,
    ):
        self._mask_type = mask_type
        self._node_type = node_type
        self._mask_parameters = mask_parameters
        self._node_parameters = node_parameters
        self._record_is_tuple = record_is_tuple
        self._record_is_scalar = record_is_scalar
        self._is_nonnullable_nulltype = is_nonnullable_nulltype
        self._option_type = option_type
        super().__init__(storage_type, "awkward")

    def __str__(self):
        return "ak:" + str(self.storage_type)

    def __repr__(self):
        return f"awkward<{self.storage_type!r}>"

    @property
    def mask_type(self):
        return self._mask_type

    @property
    def node_type(self):
        return self._node_type

    @property
    def mask_parameters(self):
        return self._mask_parameters

    @property
    def node_parameters(self):
        return self._node_parameters

    @property
    def record_is_tuple(self):
        return self._record_is_tuple

    @property
    def record_is_scalar(self):
        return self._record_is_scalar

    def __arrow_ext_class__(self):
        return AwkwardArrowArray

    def __arrow_ext_serialize__(self):
        return json.dumps(self._metadata_as_dict()).encode(errors="surrogatescape")

    def _metadata_as_dict(self):
        return {
            "mask_type": self._mask_type,
            "node_type": self._node_type,
            "mask_parameters": self._mask_parameters,
            "node_parameters": self._node_parameters,
            "record_is_tuple": self._record_is_tuple,
            "record_is_scalar": self._record_is_scalar,
            "is_nonnullable_nulltype": self._is_nonnullable_nulltype,
            "option_type": self._option_type,
        }

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        # pyarrow calls this internally
        metadata = json.loads(serialized.decode(errors="surrogatescape"))
        return cls._from_metadata_object(storage_type, metadata)

    @classmethod
    def _from_metadata_object(cls, storage_type, metadata):
        return cls(
            storage_type,
            metadata["mask_type"],
            metadata["node_type"],
            metadata["mask_parameters"],
            metadata["node_parameters"],
            metadata["record_is_tuple"],
            metadata["record_is_scalar"],
            is_nonnullable_nulltype=metadata.get("is_nonnullable_nulltype", False),
            option_type=metadata.get("option_type", False),
        )

    @property
    def num_buffers(self):
        return self.storage_type.num_buffers

    @property
    def num_fields(self):
        return self.storage_type.num_fields

    def field(self, i: int | str):
        return self.storage_type.field(i)


pyarrow.register_extension_type(
    AwkwardArrowType(pyarrow.null(), None, None, None, None, None, None)
)


def to_awkwardarrow_storage_types(arrowtype):
    if isinstance(arrowtype, AwkwardArrowType):
        return arrowtype, arrowtype.storage_type
    else:
        return None, arrowtype


def get_field_option(field: pyarrow.Field, key: bytes) -> bool | None:
    """
    Retrieves the metadata value for a given key from a field's metadata,
    interpreting the value as a boolean if it's b'True' or b'False'.
    Returns None if the key is missing.
    """
    meta = field.metadata
    if not meta:
        return None

    value = meta.get(key)
    if value is None:
        return None

    if value == b"False":
        return False
    elif value == b"True":
        return True
    else:
        return None



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from functools import cached_property

from awkward._meta.meta import Meta
from awkward._nplikes.shape import ShapeItem
from awkward._typing import JSONSerializable


class NumpyMeta(Meta):
    is_numpy = True
    is_leaf = True
    inner_shape: tuple[ShapeItem, ...]

    def purelist_parameters(self, *keys: str) -> JSONSerializable:
        if self._parameters is not None:
            for key in keys:
                if key in self._parameters:
                    return self._parameters[key]
        return None

    @property
    def purelist_isregular(self) -> bool:
        return True

    @property
    def purelist_depth(self) -> int:
        return len(self.inner_shape) + 1

    @property
    def is_identity_like(self) -> bool:
        return False

    @cached_property
    def minmax_depth(self) -> tuple[int, int]:  # type: ignore[override]
        depth = len(self.inner_shape) + 1
        return (depth, depth)

    @cached_property
    def branch_depth(self) -> tuple[bool, int]:  # type: ignore[override]
        return (False, len(self.inner_shape) + 1)

    @property
    def fields(self) -> list[str]:
        return []

    @property
    def is_tuple(self) -> bool:
        return False

    @property
    def dimension_optiontype(self) -> bool:
        return False



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function

__all__ = ("from_torch",)


@high_level_function()
def from_torch(array):
    """
    Args:
        array: (PyTorch Tensor):
            Tensor to convert into an Awkward Array.

    Converts a PyTorch Tensor into an Awkward Array.

    If `array` contains any other data types the function raises an error.
    """

    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array)


def _impl(array):
    try:
        import torch
    except ImportError as err:
        raise ImportError(
            """to use ak.from_torch, you must install 'torch' package with:

         pip install torch

or

        conda install pytorch"""
        ) from err

    # check if array is a Tensor
    if not isinstance(array, torch.Tensor):
        raise TypeError("""only PyTorch Tensor can be converted to Awkward Array""")

    # keep the resulting array on the same device as input tensor
    device = "cuda" if array.is_cuda else "cpu"

    # convert tensors to cupy if they are on cuda
    if device == "cuda":
        from awkward._nplikes.cupy import Cupy

        cp = Cupy.instance()

        # zero-copy data exchange through DLPack
        cp_array = cp.from_dlpack(array)
        ak_array = ak.from_cupy(cp_array)

    else:
        np_array = array.numpy()
        ak_array = ak.from_numpy(np_array)

    return ak_array



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward._typing import (
    ClassVar,
    JSONMapping,
    JSONSerializable,
    Self,
)
from awkward._util import UNSET, Sentinel


class Meta:
    is_numpy: ClassVar[bool] = False
    is_unknown: ClassVar[bool] = False
    is_list: ClassVar[bool] = False
    is_regular: ClassVar[bool] = False
    is_option: ClassVar[bool] = False
    is_indexed: ClassVar[bool] = False
    is_record: ClassVar[bool] = False
    is_union: ClassVar[bool] = False
    is_leaf: ClassVar[bool] = False
    is_unmasked: ClassVar[bool] = False

    _parameters: JSONMapping | None

    @property
    def parameters(self) -> JSONMapping:
        """
        Free-form parameters associated with every array node as a dict from parameter
        name to its JSON-like value. Some parameters are special and are used to assign
        behaviors to the data.

        Note that the dict returned by this property is a *view* of the array node's
        parameters. *Changing the dict will change the array!*

        See #ak.behavior.
        """
        if self._parameters is None:
            self._parameters = {}
        return self._parameters

    def parameter(self, key: str) -> JSONSerializable:
        """
        Returns a parameter's value or None.

        (No distinction is ever made between unset parameters and parameters set to None.)
        """
        if self._parameters is None:
            return None
        else:
            return self._parameters.get(key)

    def purelist_parameter(self, key: str) -> JSONSerializable:
        """
        Return the value of the outermost parameter matching `key` in a sequence
        of nested lists, stopping at the first record or tuple layer.

        If a layer has #ak.types.UnionType, the value is only returned if all
        possibilities have the same value.
        """
        return self.purelist_parameters(key)

    def purelist_parameters(self, *keys: str) -> JSONSerializable:
        """
        Return the value of the outermost parameter matching one of `keys` in a sequence
        of nested lists, stopping at the first record or tuple layer.

        If a layer has #ak.types.UnionType, the value is only returned if all
        possibilities have the same value.
        """
        raise NotImplementedError

    @property
    def is_identity_like(self) -> bool:
        raise NotImplementedError

    @property
    def purelist_isregular(self) -> bool:
        """
        Returns True if all dimensions down to the first record or tuple layer have
        #ak.types.RegularType; False otherwise.
        """
        raise NotImplementedError

    @property
    def purelist_depth(self) -> int:
        """
        Number of dimensions of nested lists, not counting anything deeper than the
        first record or tuple layer, if any. The depth of a one-dimensional array is
        `1`.

        If the array contains #ak.types.UnionType data and its contents have
        equal depths, the return value is that depth. If they do not have equal
        depths, the return value is `-1`.
        """
        raise NotImplementedError

    @property
    def minmax_depth(self) -> tuple[int, int]:
        raise NotImplementedError

    @property
    def branch_depth(self) -> tuple[bool, int]:
        raise NotImplementedError

    @property
    def fields(self) -> list[str]:
        raise NotImplementedError

    @property
    def is_tuple(self) -> bool:
        raise NotImplementedError

    @property
    def dimension_optiontype(self) -> bool:
        raise NotImplementedError

    def copy(self, *, parameters: JSONMapping | None | Sentinel = UNSET) -> Self:
        raise NotImplementedError



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import base64
import os
import struct
import sys
from collections.abc import Collection

import numpy as np  # noqa: TID251
import packaging.version

import awkward as ak
from awkward._typing import TypeVar

win = os.name == "nt"
bits32 = struct.calcsize("P") * 8 == 32
numpy2 = packaging.version.parse(np.__version__) >= packaging.version.Version("2.0.0b1")


# matches include/awkward/common.h
kMaxInt8 = 127  # 2**7  - 1
kMaxUInt8 = 255  # 2**8  - 1
kMaxInt32 = 2147483647  # 2**31 - 1
kMaxUInt32 = 4294967295  # 2**32 - 1
kMaxInt64 = 9223372036854775806  # 2**63 - 2: see below
kSliceNone = kMaxInt64 + 1  # for Slice::none()
kMaxLevels = 48


def in_module(obj, modulename: str) -> bool:
    m = type(obj).__module__
    return m == modulename or m.startswith(modulename + ".")


def tobytes(array):
    if hasattr(array, "tobytes"):
        return array.tobytes()
    else:
        return array.tostring()


native_byteorder = "<" if sys.byteorder == "little" else ">"


def native_to_byteorder(array, byteorder: str):
    """
    Args:
        array: nplike array
        byteorder (`"<"` or `">"`): desired byteorder

    Return a copy of array. Swap the byteorder if `byteorder` does not match
    `ak._util.native_byteorder`. This function is _not_ idempotent; no metadata
    from `array` exists to determine its current byteorder.
    """
    assert byteorder in "<>"
    if byteorder != native_byteorder:
        return array.byteswap(inplace=False)
    else:
        return array


def identifier_hash(str):
    return (
        base64.encodebytes(struct.pack("q", hash(str)))
        .rstrip(b"=\n")
        .replace(b"+", b"")
        .replace(b"/", b"")
        .decode("ascii")
    )


class Sentinel:
    """A class for implementing sentinel types"""

    def __init__(self, name, module=None):
        self._name = name
        self._module = module

    def __repr__(self):
        if self._module is not None:
            return f"{self._module}.{self._name}"
        else:
            return f"{self._name}"


UNSET = Sentinel("UNSET", __name__)

STDOUT = Sentinel("STDOUT", __name__)
STDOUT.stream = sys.stdout


T = TypeVar("T")


def unique_list(items: Collection[T]) -> list[T]:
    seen = set()
    result = []
    for item in items:
        if item in seen:
            continue
        seen.add(item)
        result.append(item)
    return result


def copy_behaviors(from_name: str, to_name: str, behavior: dict):
    output = {}

    for key, value in behavior.items():
        if isinstance(key, str):
            if key == from_name:
                output[to_name] = value
        else:
            if from_name in key:
                new_tuple = tuple(to_name if k == from_name else k for k in key)
                output[new_tuple] = value

    return output


def maybe_shape_of(
    obj: ak._nplikes.ArrayLike,
) -> tuple[ak._nplikes.shape.ShapeItem, ...]:
    """
    Gets the shape of an object while keeping `unknown_length` in the shape tuple.
    Virtual arrays `.shape` property materializes the shape so we use this function
    to get the shape of objects without materializing it in the case of virtual arrays.
    Unknown dimensions will be represted as `unknown_length`.
    """
    if isinstance(obj, ak._nplikes.virtual.VirtualNDArray):
        return obj._shape
    else:
        return obj.shape


def maybe_length_of(
    obj: ak.contents.Content | ak.index.Index,
) -> ak._nplikes.shape.ShapeItem:
    """
    Gets the length of an object if it is known, otherwise returns `unknown_length`.
    We use this function to get the length of objects without materializing the underlying
    virtual array buffers' shape.
    Useful for example in `__init__` methods and `__repr__` where we don't want to
    materialize the virtual array shapes.
    """
    if isinstance(obj, (ak.contents.NumpyArray, ak.index.Index)):
        return maybe_shape_of(obj._data)[0]
    elif isinstance(obj, ak.contents.ListOffsetArray):
        return maybe_length_of(obj._offsets) - 1
    elif isinstance(obj, ak.contents.ListArray):
        return maybe_length_of(obj._starts)
    elif isinstance(
        obj,
        (ak.contents.RecordArray, ak.contents.RegularArray, ak.contents.BitMaskedArray),
    ):
        return obj._length
    elif isinstance(obj, ak.contents.ByteMaskedArray):
        return maybe_length_of(obj._mask)
    elif isinstance(obj, (ak.contents.IndexedArray, ak.contents.IndexedOptionArray)):
        return maybe_length_of(obj._index)
    elif isinstance(obj, ak.contents.UnionArray):
        return maybe_length_of(obj._tags)
    elif isinstance(obj, ak.contents.UnmaskedArray):
        return maybe_length_of(obj._content)
    elif isinstance(obj, ak.contents.EmptyArray):
        return 0
    else:
        raise TypeError(f"Invalid type {type(obj)} for length calculation")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy

__all__ = ("AxisError", "FieldNotFoundError")


class FieldNotFoundError(IndexError):
    pass


AxisError = getattr(numpy, "exceptions", numpy).AxisError



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward._backends.backend import Backend
from awkward.contents.content import Content
from awkward.record import Record


# Note: remove_structure is semi-public, exposed via awkward.contents.content.
def remove_structure(
    layout: Content | Record,
    backend: Backend | None = None,
    flatten_records: bool = True,
    function_name: str | None = None,
    drop_nones: bool = True,
    keepdims: bool = False,
    allow_records: bool = False,
    list_to_regular: bool = False,
):
    if isinstance(layout, Record):
        return remove_structure(
            layout._array[layout._at : layout._at + 1],
            backend,
            flatten_records,
            function_name,
            drop_nones,
            keepdims,
            allow_records,
        )

    else:
        if backend is None:
            backend = layout._backend
        arrays = layout._remove_structure(
            backend,
            {
                "flatten_records": flatten_records,
                "function_name": function_name,
                "drop_nones": drop_nones,
                "keepdims": keepdims,
                "allow_records": allow_records,
                "list_to_regular": list_to_regular,
            },
        )
        return tuple(arrays)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._attrs import attrs_of_obj
from awkward._dispatch import high_level_function
from awkward._layout import (
    HighLevelContext,
    ensure_same_backend,
    maybe_highlevel_to_lowlevel,
)
from awkward._namedaxis import (
    AxisName,
    _get_named_axis,
)
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._typing import Mapping

__all__ = ("moment",)

np = NumpyMetadata.instance()


@high_level_function()
def moment(
    x,
    n,
    weight=None,
    axis=None,
    *,
    keepdims: bool = False,
    mask_identity: bool = False,
    highlevel: bool = True,
    behavior: Mapping | None = None,
    attrs: Mapping | None = None,
):
    """
    Args:
        x: The data on which to compute the moment (anything #ak.to_layout recognizes).
        n (int): The choice of moment: `0` is a sum of weights, `1` is
            #ak.mean, `2` is #ak.var without subtracting the mean, etc.
        weight: Data that can be broadcasted to `x` to give each value a
            weight. Weighting values equally is the same as no weights;
            weighting some values higher increases the significance of those
            values. Weights can be zero or negative.
        axis (None or int or str): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc; if a str, it is interpreted as the
            name of the axis which maps to an int if named axes are present.
            Named axes are attached to an array using #ak.with_named_axis and
            removed with #ak.without_named_axis; also see the
            [Named axes user guide](../../user-guide/how-to-array-properties-named-axis.html).
        keepdims (bool): If False, this function decreases the number of
            dimensions by 1; if True, the output values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        mask_identity (bool): If True, the application of this function on
            empty lists results in None (an option type); otherwise, the
            calculation is followed through with the reducers' identities,
            usually resulting in floating-point `nan`.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Computes the `n`th moment in each group of elements from `x` (many
    types supported, including all Awkward Arrays and Records). The grouping
    is performed the same way as for reducers, though this operation is not a
    reducer and has no identity.

    This function has no NumPy equivalent.

    Passing all arguments to the reducers, the moment is calculated as

        ak.sum(x**n * weight) / ak.sum(weight)

    The `n=2` moment differs from #ak.var in that #ak.var also subtracts the
    mean (the `n=1` moment).

    See #ak.sum for a complete description of handling nested lists and
    missing values (None) in reducers, and #ak.mean for an example with another
    non-reducer.
    """
    # Dispatch
    yield x, weight

    # Implementation
    return _impl(
        x, n, weight, axis, keepdims, mask_identity, highlevel, behavior, attrs
    )


def _impl(
    x,
    n,
    weight,
    axis: AxisName,
    keepdims: bool,
    mask_identity: bool,
    highlevel: bool,
    behavior: Mapping | None,
    attrs: Mapping | None,
):
    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        x_layout, weight_layout = ensure_same_backend(
            ctx.unwrap(x, allow_record=False, primitive_policy="error"),
            ctx.unwrap(
                weight,
                allow_record=False,
                allow_unknown=False,
                primitive_policy="error",
                none_policy="pass-through",
            ),
        )

    x = ctx.wrap(x_layout)
    weight = ctx.wrap(weight_layout, allow_other=True)

    with np.errstate(invalid="ignore", divide="ignore"):
        if weight is None:
            sumw = ak.operations.ak_count._impl(
                x,
                axis,
                keepdims,
                mask_identity,
                highlevel=True,
                behavior=ctx.behavior,
                attrs=ctx.attrs,
            )
            sumwxn = ak.operations.ak_sum._impl(
                x**n,
                axis,
                keepdims,
                mask_identity,
                highlevel=True,
                behavior=ctx.behavior,
                attrs=ctx.attrs,
            )
        else:
            sumw = ak.operations.ak_sum._impl(
                x * 0 + weight,
                axis,
                keepdims,
                mask_identity,
                highlevel=True,
                behavior=ctx.behavior,
                attrs=ctx.attrs,
            )
            sumwxn = ak.operations.ak_sum._impl(
                (x**n) * weight,
                axis,
                keepdims,
                mask_identity,
                highlevel=True,
                behavior=ctx.behavior,
                attrs=ctx.attrs,
            )

        out = sumwxn / sumw

        # propagate named axis to output
        wrapped = ctx.wrap(
            maybe_highlevel_to_lowlevel(out),
            highlevel=highlevel,
            allow_other=True,
        )
        # propagate named axis to output
        return ak.operations.ak_with_named_axis._impl(
            wrapped,
            named_axis=_get_named_axis(attrs_of_obj(out), allow_any=True),
            highlevel=highlevel,
            behavior=None,
            attrs=None,
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import jax

import awkward as ak
from awkward import contents, highlevel, record
from awkward._backends.jax import JaxBackend
from awkward._layout import HighLevelContext
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._typing import Generic, TypeVar
from awkward.contents import Content

T = TypeVar(
    "T", bound=highlevel.Array | highlevel.Record | contents.Content | record.Record
)

np = NumpyMetadata.instance()


def split_buffers(buffers: dict) -> tuple[dict, dict]:
    data_buffers, other_buffers = {}, {}
    for key, buf in buffers.items():
        _, attr = key.rsplit("-", 1)
        if attr == "data":
            data_buffers[key] = buf
        else:
            other_buffers[key] = buf
    return data_buffers, other_buffers


class AuxData(Generic[T]):
    """
    This class is used to store the auxiliary data needed to reconstruct an Awkward Array from its buffers.

    AuxData deliberately can not use the layout, because the layout has a reference to the buffers which are replaced
    by JAX with tracers - this leads to tracer leaks!

    Instead, this class holds the form, length, and other numpy buffers needed to reconstruct it.
    """

    def __init__(
        self,
        data_buffer_keys: tuple[str],
        other_buffers: dict,
        form: ak.forms.Form,
        length: int,
        ctx: HighLevelContext,
        highlevel: bool = True,
    ):
        self._data_buffer_keys = data_buffer_keys
        self._other_buffers = other_buffers
        self._form = form
        self._length = length
        self._ctx = ctx
        self._highlevel = highlevel

    @classmethod
    def from_array_or_layout(cls, obj: T):
        with HighLevelContext() as ctx:
            layout = ctx.unwrap(obj, allow_record=True, primitive_policy="error")

        # change backend
        jax_backend = JaxBackend.instance()
        layout = layout.to_backend(jax_backend)

        # decompose
        form, length, buffers = ak.operations.to_buffers(layout)

        # we need to split buffers into all "data" buffers and all others (e.g. index, offsets, etc.)
        data_buffers, other_buffers = split_buffers(buffers)

        # now we need to flatten the data buffers
        data_buffer_keys, data_flat_buffers = zip(*data_buffers.items(), strict=True)
        return data_flat_buffers, AuxData(
            data_buffer_keys=data_buffer_keys,
            other_buffers=other_buffers,
            form=form,
            length=length,
            ctx=ctx,
            highlevel=not isinstance(obj, Content),
        )

    def unflatten(self, data_buffers: tuple) -> T:
        for buffer in data_buffers:
            # Check that JAX isn't trying to give us float0 types
            dtype = getattr(buffer, "dtype", None)
            if dtype == np.dtype([("float0", "V")]):
                raise TypeError(
                    f"a buffer with the dtype {buffer.dtype} was encountered during unflattening. "
                    "JAX uses this dtype for the tangents of integer/boolean outputs; these cannot "
                    "reasonably be differentiated. Make sure that you are not computing the derivative "
                    "of a boolean/integer (array) valued function."
                )

        # reconstitute data buffers
        data_buffers = dict(zip(self._data_buffer_keys, data_buffers, strict=True))

        # combine data buffers with other buffers
        buffers = {**self._other_buffers, **data_buffers}

        # reconstruct layout
        layout = ak.operations.from_buffers(
            self._form,
            self._length,
            buffers,
            backend="jax",
            highlevel=False,  # we will wrap it later
        )

        # wrap layout
        return self._ctx.wrap(
            layout,
            highlevel=self._highlevel,
            allow_other=True,
        )

    def __eq__(self, other: AuxData) -> bool:
        return (
            self._form.is_equal_to(
                other=other._form, all_parameters=True, form_key=True
            )
            and self._length == other._length
        )

    def __ne__(self, other: AuxData) -> bool:
        return not self == other


def jax_flatten(
    array: T,
) -> tuple[tuple, AuxData]:
    result = AuxData.from_array_or_layout(array)
    return result


def jax_unflatten(aux_data: AuxData, children: tuple) -> T | None:
    return aux_data.unflatten(children)


HighLevelType = TypeVar(
    "HighLevelType", bound="type[highlevel.Array | highlevel.Record]"
)


def register_pytree_class(cls: T) -> T:
    """
    Args:
        cls: class to register with JAX

    Return the class, after registering it with JAX.

    """
    jax.tree_util.register_pytree_node(
        cls,
        jax_flatten,
        jax_unflatten,
    )
    return cls



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import math
from functools import lru_cache, partial

import awkward as ak
from awkward._backends.dispatch import regularize_backend
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.array_like import ArrayLike
from awkward._nplikes.jax import Jax
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpy_like import NumpyLike, NumpyMetadata
from awkward._nplikes.placeholder import PlaceholderArray
from awkward._nplikes.shape import ShapeItem, unknown_length
from awkward._nplikes.virtual import VirtualNDArray
from awkward._regularize import is_integer
from awkward._typing import Callable
from awkward.forms.form import index_to_dtype, regularize_buffer_key

__all__ = ("from_buffers",)

np = NumpyMetadata.instance()
numpy = Numpy.instance()


@high_level_function()
def from_buffers(
    form,
    length,
    container,
    buffer_key="{form_key}-{attribute}",
    *,
    backend="cpu",
    byteorder="<",
    allow_noncanonical_form=False,
    enable_virtualarray_caching=True,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        form (#ak.forms.Form or str/dict equivalent): The form of the Awkward
            Array to reconstitute from named buffers.
        length (int): Length of the array. (The output of this function is always
            single-partition.)
        container (Mapping, such as dict): The str \u2192 Python buffers that
            represent the decomposed Awkward Array. This `container` is only
            assumed to have a `__getitem__` method that accepts strings as keys.
        buffer_key (str or callable): Python format string containing
            `"{form_key}"` and/or `"{attribute}"` or a function that takes these
            as keyword arguments and returns a string to use as a key for a buffer
            in the `container`.
        backend (str): Library to use to generate values that are
            put into the new array. The default, cpu, makes NumPy
            arrays, which are in main memory (e.g. not GPU). If all the values in
            `container` have the same `backend` as this, they won't be copied.
        byteorder (`"<"`, `">"`): Endianness of buffers read from `container`.
            If the byteorder does not match the current system byteorder, the
            arrays will be copied.
        allow_noncanonical_form (bool): If True, non-canonical forms will be
            simplified to produce arrays with canonical layouts; otherwise,
            an exception will be thrown for such forms.
        enable_virtualarray_caching (bool or callable): If True (the default),
            all VirtualNDArray buffers that get created will cache their
            materialized buffers on themselves when they get materialized.
            If a callable is given, it must accept two arguments, `form_key` and `attribute`,
            and return a boolean indicating whether caching should be enabled for the
            given buffer. The `form_key` and `attribute` are the same as those passed to
            the `buffer_key` function. If False, all VirtualNDArrays will not cache
            their materialized buffers on themselves.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Reconstitutes an Awkward Array from a Form, length, and a collection of memory
    buffers, so that data can be losslessly read from file formats and storage
    devices that only map names to binary blobs (such as a filesystem directory).

    The first three arguments of this function are the return values of
    #ak.to_buffers, so a full round-trip is

        >>> reconstituted = ak.from_buffers(*ak.to_buffers(original))

    The `container` argument lets you specify your own Mapping, which might be
    an interface to some storage format or device (e.g. h5py). It's okay if
    the `container` dropped NumPy's `dtype` and `shape` information, leaving
    raw bytes, since `dtype` and `shape` can be reconstituted from the
    #ak.forms.NumpyForm.
    If the values of `container` are recognised as arrays by the given backend,
    a view over their existing data will be used, where possible.
    The `container` values are allowed to be callables with no arguments.
    If that's the case, they will be turned into `VirtualNDArray` buffers whose generator
    function is the callable and is used to materialize the buffer when required.

    The `buffer_key` should be the same as the one used in #ak.to_buffers.

    When `allow_noncanonical_form` is set to True, this function readily accepts
    non-simplified forms, i.e. forms which will be simplified by Awkward Array
    into "canonical" representations, e.g. `option[option[...]]` → `option[...]`.
    Such forms can be produced by the low-level ArrayBuilder `snapshot()` method.
    Given that Awkward Arrays must have canonical layouts, it follows that
    invoking this function with `allow_noncanonical_form` may produce arrays
    whose forms differ to the input form.

    In order for a non-simplified form to be considered valid, it should be one
    that the #ak.contents.Content layout classes could produce iff. the
    simplification rules were removed.


    See #ak.to_buffers for examples.
    """
    return _impl(
        form,
        length,
        container,
        buffer_key,
        backend,
        byteorder,
        highlevel,
        behavior,
        attrs,
        allow_noncanonical_form,
        enable_virtualarray_caching,
    )


def _impl(
    form,
    length,
    container,
    buffer_key,
    backend,
    byteorder,
    highlevel,
    behavior,
    attrs,
    simplify,
    enable_virtualarray_caching,
):
    backend = regularize_backend(backend)
    if not backend.nplike.known_data:
        msg = (
            "Typetacer backend is not supported in 'ak.from_buffers'. "
            "Typetracer-backed arrays can be constructed only from the form as they do not hold any data. "
            "Use highlevel functions like 'ak.typetracer.typetracer_from_form' or 'ak.typetracer.typetracer_with_report' "
            "to construct such arrays."
        )
        raise TypeError(msg)

    if isinstance(form, str):
        if ak.types.numpytype.is_primitive(form):
            form = ak.forms.NumpyForm(form)
        else:
            form = ak.forms.from_json(form)
    elif isinstance(form, dict):
        form = ak.forms.from_dict(form)

    if not (is_integer(length) and length >= 0):
        raise TypeError("'length' argument must be a non-negative integer")
    length = int(length)

    if not isinstance(form, ak.forms.Form):
        raise TypeError(
            "'form' argument must be a Form or its Python dict/JSON string representation"
        )

    if isinstance(enable_virtualarray_caching, bool):

        def enable_caching_function(form_key, attribute):
            return enable_virtualarray_caching
    elif callable(enable_virtualarray_caching):
        enable_caching_function = enable_virtualarray_caching
    else:
        raise TypeError(
            "'enable_virtualarray_caching' argument must be a boolean or a callable"
        )

    getkey = regularize_buffer_key(buffer_key)

    out = _reconstitute(
        form,
        length,
        container,
        getkey,
        backend,
        byteorder,
        simplify,
        shape_generator=lambda: (length,),
        enable_virtualarray_caching=enable_caching_function,
    )

    return wrap_layout(out, highlevel=highlevel, attrs=attrs, behavior=behavior)


def _from_buffer(
    nplike: NumpyLike,
    buffer: Callable | ArrayLike,
    dtype: np.dtype,
    count: ShapeItem,
    byteorder: str,
    buffer_key: str,
    shape_generator: Callable | None = None,
    enable_virtualarray_caching: bool = False,
) -> ArrayLike:
    if isinstance(buffer, VirtualNDArray):
        # This is the case for VirtualNDArrays
        # just some checks to make sure the VirtualNDArray is correctly constructed
        if nplike != buffer.nplike:
            raise ValueError(
                f"Mismatch of nplikes. Got {nplike}, but VirtualNDArray has {buffer.nplike}."
            )
        if dtype != buffer.dtype:
            raise ValueError(
                f"Mismatch of dtypes. Got {dtype}, but VirtualNDArray has {buffer.dtype}."
            )
        if (
            count is not unknown_length
            and buffer._shape[0] is not unknown_length
            and count != buffer._shape[0]
        ):
            raise ValueError(
                f"Mismatch of lengths. Got {count}, but VirtualNDArray has {buffer._shape[0]}."
            )
        return buffer

    elif callable(buffer):
        # This is the case where we automatically create VirtualNDArrays
        # We use recursion here to pass down the from_buffer and byteorder transformations to the generator
        assert callable(shape_generator), "shape_generator must be callable"
        cached_shape_generator = lru_cache(maxsize=1)(shape_generator)

        def generator():
            (length,) = cached_shape_generator()
            return _from_buffer(
                nplike,
                buffer(),
                dtype,
                length,
                byteorder,
                buffer_key,
                None,
                False,
            )

        # also store a ref to the original/raw buffer generator
        # this allows us to access it later again
        generator.__awkward_raw_generator__ = buffer

        return VirtualNDArray(
            nplike=nplike,
            shape=(count,),
            dtype=dtype,
            generator=generator,
            shape_generator=cached_shape_generator,
            buffer_key=buffer_key,
            __wrap_generator_asarray__=True,
            __enable_caching__=enable_virtualarray_caching,
        )
    # Unknown-length information implies that we didn't load shape-buffers (offsets, etc)
    # for the parent of this node. Thus, this node and its children *must* only
    # contain placeholders
    elif count is unknown_length:
        # We may actually have a known buffer here, but as we do not know the length,
        # we cannot safely trim it. Thus, introduce a placeholder anyway
        return PlaceholderArray(nplike, (unknown_length,), dtype, buffer_key)
    # Known-length information implies that we should have known-length buffers here
    # We could choose to make this an error, and have the caller re-implement some
    # of #ak.from_buffers, or we can just introduce the known lengths where possible
    elif isinstance(buffer, PlaceholderArray) and buffer.size is unknown_length:
        return PlaceholderArray(nplike, (count,), dtype, buffer_key)
    elif isinstance(buffer, PlaceholderArray) or nplike.is_own_array(buffer):
        # Require 1D buffers
        copy = None if isinstance(nplike, Jax) else False  # Jax can not avoid this
        array = nplike.reshape(buffer.view(dtype), shape=(-1,), copy=copy)
        array = ak._util.native_to_byteorder(array, byteorder)

        # we can't compare with count or slice when we're working with tracers
        if not (isinstance(nplike, Jax) and nplike.is_currently_tracing()):
            if array.size < count:
                raise TypeError(
                    f"size of array ({array.size}) is less than size of form ({count})"
                )
            return array[:count]
        else:
            return array
    else:
        array = nplike.frombuffer(buffer, dtype=dtype, count=count)
        return ak._util.native_to_byteorder(array, byteorder)


def _reconstitute(
    form,
    length,
    container,
    getkey,
    backend,
    byteorder,
    simplify,
    shape_generator,
    enable_virtualarray_caching,
):
    if isinstance(form, ak.forms.EmptyForm):
        if length is not unknown_length and length != 0:
            raise ValueError(f"EmptyForm node, but the expected length is {length}")
        return ak.contents.EmptyArray(backend=backend)

    elif isinstance(form, ak.forms.NumpyForm):
        dtype = ak.types.numpytype.primitive_to_dtype(form.primitive)
        raw_array = container[(key := getkey(form, "data"))]

        def _adjust_length(length):
            return length * math.prod(form.inner_shape)

        real_length = _adjust_length(length)

        def _shape_generator():
            (length,) = shape_generator()
            return (_adjust_length(length),)

        data = _from_buffer(
            backend.nplike,
            raw_array,
            dtype=dtype,
            count=real_length,
            byteorder=byteorder,
            buffer_key=key,
            shape_generator=_shape_generator,
            enable_virtualarray_caching=enable_virtualarray_caching(
                form.form_key, "data"
            ),
        )
        if form.inner_shape != ():
            data = backend.nplike.reshape(data, (length, *form.inner_shape))

        return ak.contents.NumpyArray(
            data, parameters=form._parameters, backend=backend
        )

    elif isinstance(form, ak.forms.UnmaskedForm):
        content = _reconstitute(
            form.content,
            length,
            container,
            getkey,
            backend,
            byteorder,
            simplify,
            shape_generator,
            enable_virtualarray_caching,
        )
        if simplify:
            make = ak.contents.UnmaskedArray.simplified
        else:
            make = ak.contents.UnmaskedArray
        return make(content, parameters=form._parameters)

    elif isinstance(form, ak.forms.BitMaskedForm):
        raw_array = container[(key := getkey(form, "mask"))]

        def _adjust_length(length):
            return math.ceil(length / 8.0)

        def _shape_generator():
            (length,) = shape_generator()
            return (_adjust_length(length),)

        if length is unknown_length:
            next_length = unknown_length
        else:
            next_length = _adjust_length(length)

        mask = _from_buffer(
            backend.nplike,
            raw_array,
            dtype=index_to_dtype[form.mask],
            count=next_length,
            byteorder=byteorder,
            buffer_key=key,
            shape_generator=_shape_generator,
            enable_virtualarray_caching=enable_virtualarray_caching(
                form.form_key, "mask"
            ),
        )
        content = _reconstitute(
            form.content,
            length,
            container,
            getkey,
            backend,
            byteorder,
            simplify,
            shape_generator,
            enable_virtualarray_caching,
        )
        if simplify:
            make = ak.contents.BitMaskedArray.simplified
        else:
            make = ak.contents.BitMaskedArray
        # We need to know the length of a BitMaskedArray to initialize it
        # as it is an argument in __init__ and is not calculated from the content
        (length,) = shape_generator()
        return make(
            ak.index.Index(mask),
            content,
            form.valid_when,
            length,
            form.lsb_order,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.ByteMaskedForm):
        raw_array = container[(key := getkey(form, "mask"))]
        mask = _from_buffer(
            backend.nplike,
            raw_array,
            dtype=index_to_dtype[form.mask],
            count=length,
            byteorder=byteorder,
            buffer_key=key,
            shape_generator=shape_generator,
            enable_virtualarray_caching=enable_virtualarray_caching(
                form.form_key, "mask"
            ),
        )
        content = _reconstitute(
            form.content,
            length,
            container,
            getkey,
            backend,
            byteorder,
            simplify,
            shape_generator,
            enable_virtualarray_caching,
        )
        if simplify:
            make = ak.contents.ByteMaskedArray.simplified
        else:
            make = ak.contents.ByteMaskedArray
        return make(
            ak.index.Index(mask),
            content,
            form.valid_when,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.IndexedOptionForm):
        raw_array = container[(key := getkey(form, "index"))]
        index = _from_buffer(
            backend.nplike,
            raw_array,
            dtype=index_to_dtype[form.index],
            count=length,
            byteorder=byteorder,
            buffer_key=key,
            shape_generator=shape_generator,
            enable_virtualarray_caching=enable_virtualarray_caching(
                form.form_key, "index"
            ),
        )

        def _adjust_length(index):
            return 0 if len(index) == 0 else max(0, backend.nplike.max(index) + 1)

        def _shape_generator():
            return (_adjust_length(index),)

        if isinstance(index, (PlaceholderArray, VirtualNDArray)):
            next_length = unknown_length
        else:
            next_length = _adjust_length(index)
        content = _reconstitute(
            form.content,
            next_length,
            container,
            getkey,
            backend,
            byteorder,
            simplify,
            _shape_generator,
            enable_virtualarray_caching,
        )
        if simplify:
            make = ak.contents.IndexedOptionArray.simplified
        else:
            make = ak.contents.IndexedOptionArray
        return make(
            ak.index.Index(index),
            content,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.IndexedForm):
        raw_array = container[(key := getkey(form, "index"))]
        index = _from_buffer(
            backend.nplike,
            raw_array,
            dtype=index_to_dtype[form.index],
            count=length,
            byteorder=byteorder,
            buffer_key=key,
            shape_generator=shape_generator,
            enable_virtualarray_caching=enable_virtualarray_caching(
                form.form_key, "index"
            ),
        )

        def _adjust_length(index):
            return (
                0
                if len(index) == 0
                else backend.nplike.index_as_shape_item(backend.nplike.max(index) + 1)
            )

        def _shape_generator():
            return (_adjust_length(index),)

        if isinstance(index, (PlaceholderArray, VirtualNDArray)):
            next_length = unknown_length
        else:
            next_length = _adjust_length(index)
        content = _reconstitute(
            form.content,
            next_length,
            container,
            getkey,
            backend,
            byteorder,
            simplify,
            _shape_generator,
            enable_virtualarray_caching,
        )
        if simplify:
            make = ak.contents.IndexedArray.simplified
        else:
            make = ak.contents.IndexedArray
        return make(
            ak.index.Index(index),
            content,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.ListForm):
        raw_array1 = container[(key1 := getkey(form, "starts"))]
        raw_array2 = container[(key2 := getkey(form, "stops"))]
        starts = _from_buffer(
            backend.nplike,
            raw_array1,
            dtype=index_to_dtype[form.starts],
            count=length,
            byteorder=byteorder,
            buffer_key=key1,
            shape_generator=shape_generator,
            enable_virtualarray_caching=enable_virtualarray_caching(
                form.form_key, "starts"
            ),
        )
        stops = _from_buffer(
            backend.nplike,
            raw_array2,
            dtype=index_to_dtype[form.stops],
            count=length,
            byteorder=byteorder,
            buffer_key=key2,
            shape_generator=shape_generator,
            enable_virtualarray_caching=enable_virtualarray_caching(
                form.form_key, "stops"
            ),
        )

        def _adjust_length(starts, stops):
            reduced_stops = stops[starts != stops]
            return 0 if len(reduced_stops) == 0 else backend.nplike.max(reduced_stops)

        def _shape_generator():
            return (_adjust_length(starts, stops),)

        if isinstance(starts, (PlaceholderArray, VirtualNDArray)) or isinstance(
            stops, (PlaceholderArray, VirtualNDArray)
        ):
            next_length = unknown_length
        else:
            next_length = _adjust_length(starts, stops)
        content = _reconstitute(
            form.content,
            next_length,
            container,
            getkey,
            backend,
            byteorder,
            simplify,
            _shape_generator,
            enable_virtualarray_caching,
        )
        return ak.contents.ListArray(
            ak.index.Index(starts),
            ak.index.Index(stops),
            content,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.ListOffsetForm):
        raw_array = container[(key := getkey(form, "offsets"))]

        def _shape_generator():
            (first,) = shape_generator()
            return (first + 1,)

        offsets = _from_buffer(
            backend.nplike,
            raw_array,
            dtype=index_to_dtype[form.offsets],
            count=length + 1,
            byteorder=byteorder,
            buffer_key=key,
            shape_generator=_shape_generator,
            enable_virtualarray_caching=enable_virtualarray_caching(
                form.form_key, "offsets"
            ),
        )

        # next length
        def _adjust_length(offsets):
            return 0 if len(offsets) == 1 else offsets[-1]

        def _shape_generator():
            return (_adjust_length(offsets),)

        if isinstance(offsets, (PlaceholderArray, VirtualNDArray)):
            next_length = unknown_length
        else:
            next_length = _adjust_length(offsets)

        content = _reconstitute(
            form.content,
            next_length,
            container,
            getkey,
            backend,
            byteorder,
            simplify,
            _shape_generator,
            enable_virtualarray_caching,
        )
        return ak.contents.ListOffsetArray(
            ak.index.Index(offsets),
            content,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.RegularForm):

        def _adjust_length(length):
            return length * form.size

        next_length = _adjust_length(length)

        def _shape_generator():
            (first,) = shape_generator()
            return (_adjust_length(first),)

        content = _reconstitute(
            form.content,
            next_length,
            container,
            getkey,
            backend,
            byteorder,
            simplify,
            _shape_generator,
            enable_virtualarray_caching,
        )
        return ak.contents.RegularArray(
            content,
            form.size,
            length,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.RecordForm):
        contents = [
            _reconstitute(
                content,
                length,
                container,
                getkey,
                backend,
                byteorder,
                simplify,
                shape_generator,
                enable_virtualarray_caching,
            )
            for content, field in zip(form.contents, form.fields, strict=True)
        ]
        return ak.contents.RecordArray(
            contents,
            None if form.is_tuple else form.fields,
            length,
            parameters=form._parameters,
            backend=backend,
        )

    elif isinstance(form, ak.forms.UnionForm):
        raw_array1 = container[(key1 := getkey(form, "tags"))]
        raw_array2 = container[(key2 := getkey(form, "index"))]
        tags = _from_buffer(
            backend.nplike,
            raw_array1,
            dtype=index_to_dtype[form.tags],
            count=length,
            byteorder=byteorder,
            buffer_key=key1,
            shape_generator=shape_generator,
            enable_virtualarray_caching=enable_virtualarray_caching(
                form.form_key, "tags"
            ),
        )
        index = _from_buffer(
            backend.nplike,
            raw_array2,
            dtype=index_to_dtype[form.index],
            count=length,
            byteorder=byteorder,
            buffer_key=key2,
            shape_generator=shape_generator,
            enable_virtualarray_caching=enable_virtualarray_caching(
                form.form_key, "index"
            ),
        )

        def _adjust_length(index, tags, tag):
            selected_index = index[tags == tag]
            if len(selected_index) == 0:
                return 0
            else:
                return backend.nplike.max(selected_index) + 1

        _shape_generators = []
        for tag in range(len(form.contents)):

            def _shape_generator(tag):
                return (_adjust_length(index, tags, tag),)

            _shape_generators.append(partial(_shape_generator, tag=tag))

        if isinstance(index, (PlaceholderArray, VirtualNDArray)) or isinstance(
            tags, (PlaceholderArray, VirtualNDArray)
        ):
            lengths = [unknown_length] * len(form.contents)
        else:
            lengths = []
            for tag in range(len(form.contents)):
                lengths.append(_adjust_length(index, tags, tag))

        contents = [
            _reconstitute(
                content,
                lengths[i],
                container,
                getkey,
                backend,
                byteorder,
                simplify,
                _shape_generators[i],
                enable_virtualarray_caching,
            )
            for i, content in enumerate(form.contents)
        ]
        if simplify:
            make = ak.contents.UnionArray.simplified
        else:
            make = ak.contents.UnionArray
        return make(
            ak.index.Index(tags),
            ak.index.Index(index),
            contents,
            parameters=form._parameters,
        )

    else:
        raise AssertionError("unexpected form node type: " + str(type(form)))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import ctypes
import importlib.resources as importlib_resources
import platform

import awkward_cpp._kernel_signatures

if platform.system() == "Windows":
    name = "awkward-cpu-kernels.dll"
elif platform.system() == "Darwin":
    name = "libawkward-cpu-kernels.dylib"
else:
    name = "libawkward-cpu-kernels.so"

libpath_ref = importlib_resources.files(awkward_cpp) / "lib" / name
with importlib_resources.as_file(libpath_ref) as libpath:
    lib = ctypes.cdll.LoadLibrary(str(libpath))

kernel = awkward_cpp._kernel_signatures.by_signature(lib)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from collections import Counter
from functools import cached_property

from awkward._meta.meta import Meta
from awkward._typing import Generic, JSONSerializable, TypeVar

T = TypeVar("T", bound=Meta)


class UnionMeta(Meta, Generic[T]):
    is_union = True

    _contents: list[T]

    def purelist_parameters(self, *keys: str) -> JSONSerializable:
        if self._parameters is not None:
            for key in keys:
                if key in self._parameters:
                    return self._parameters[key]

        for key in keys:
            out = self._contents[0].purelist_parameter(key)
            for content in self._contents[1:]:
                tmp = content.purelist_parameter(key)
                if out != tmp:
                    return None
            return out

        return None

    @cached_property
    def purelist_isregular(self) -> bool:  # type: ignore[override]
        for content in self._contents:
            if not content.purelist_isregular:
                return False
        return True

    @cached_property
    def purelist_depth(self) -> int:  # type: ignore[override]
        out = None
        for content in self._contents:
            if out is None:
                out = content.purelist_depth
            elif out != content.purelist_depth:
                return -1
        assert out is not None
        return out

    @property
    def is_identity_like(self) -> bool:
        return False

    @cached_property
    def minmax_depth(self) -> tuple[int, int]:  # type: ignore[override]
        if len(self._contents) == 0:
            return (0, 0)
        mindepth, maxdepth = self._contents[0].minmax_depth
        for content in self._contents[1:]:
            mindepth_, maxdepth_ = content.minmax_depth
            if mindepth_ < mindepth:
                mindepth = mindepth_
            if maxdepth_ > maxdepth:
                maxdepth = maxdepth_
        return (mindepth, maxdepth)

    @cached_property
    def branch_depth(self) -> tuple[bool, int]:  # type: ignore[override]
        if len(self._contents) == 0:
            return False, 1

        any_branch = False
        min_depth = None
        for content in self._contents:
            branch, depth = content.branch_depth
            if min_depth is None:
                min_depth = depth
            if branch or min_depth != depth:
                any_branch = True
            if min_depth > depth:
                min_depth = depth

        assert min_depth is not None
        return any_branch, min_depth

    @cached_property
    def fields(self) -> list[str]:  # type: ignore[override]
        field_counts = Counter([f for c in self._contents for f in c.fields])
        return [f for f, n in field_counts.items() if n == len(self._contents)]

    @property
    def is_tuple(self) -> bool:
        return all(x.is_tuple for x in self._contents) and (len(self._contents) > 0)

    @property
    def dimension_optiontype(self) -> bool:
        for content in self._contents:
            if content.dimension_optiontype:
                return True
        return False

    def content(self, index: int) -> T:
        return self._contents[index]

    @cached_property
    def contents(self) -> list[T]:
        return self._contents



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE
from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function

__all__ = ("to_cudf",)


@high_level_function()
def to_cudf(array):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).

    Converts an Awkward Array into a cuDF Series.

    Buffers that are not already in GPU memory will be transferred, and some
    structural reformatting may happen to account for differences in architecture.

    This function requires the `cudf` library (< 25.12.00) and a compatible GPU.
    cuDF versions 25.12.00 and later are not currently supported due to
    incompatible changes in cuDF internals.

    See also #ak.to_cupy, #ak.from_cupy, #ak.to_dataframe.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array)


def _impl(array):
    try:
        import cudf
    except ImportError as err:
        raise ImportError(
            """to use ak.to_cudf, you must install the 'cudf' package with:

    pip install cudf-cu13
or
    conda install -c rapidsai cudf cuda-version=13"""
        ) from err

    from packaging.version import parse as parse_version

    if parse_version(cudf.__version__) >= parse_version("25.12.00"):
        raise NotImplementedError(
            f"ak.to_cudf is not supported for cudf >= 25.12.00 (you have {cudf.__version__}). "
            "cudf internals changed in ways that are incompatible with the current implementation"
        )

    layout = ak.to_layout(array, allow_record=False)

    if hasattr(cudf.Series, "_from_column"):
        return cudf.Series._from_column(layout._to_cudf(cudf, None, len(layout)))
    # older Series invocation
    return cudf.Series(layout._to_cudf(cudf, None, len(layout)))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

# Versioning
from awkward._version import __version__

# NumPy-like alternatives
from awkward import _backends
from awkward import _nplikes

# layout classes; functionality that used to be in C++ (in Awkward 1.x)
from awkward import index
from awkward import contents
from awkward import record
from awkward import types
from awkward import forms

# internal
from awkward import _do
from awkward import _slicing
from awkward import _broadcasting
from awkward import _reducers
from awkward import _util
from awkward import _errors
from awkward import _lookup
from awkward import _ext  # strictly for unpickling from Awkward 1
from awkward import _namedaxis

# third-party connectors
from awkward._connect import numpy
from awkward._connect import numexpr
from awkward import numba
from awkward import cppyy
from awkward import jax
from awkward import typetracer
from awkward import _typetracer  # todo: remove this after "deprecation" period

# high-level interface
from awkward.highlevel import *

# behaviors
from awkward.behaviors.mixins import *

# exports
from awkward import builder
from awkward import forth

# errors
from awkward import errors

behavior: dict = {}

# operations
from awkward.operations import *

# filter out private modules starting with "_" and exclude "ak_" modules from src/awkward/operations
__all__ = [
    k
    for k, v in locals().items()
    if not (
        k.startswith("_")
        or (k.startswith("ak_") and v.__package__ == "awkward.operations")
    )
]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward._nplikes.array_like import MaterializableArray, maybe_materialize
from awkward._nplikes.numpy_like import NumpyLike, NumpyMetadata
from awkward._nplikes.shape import ShapeItem, unknown_length
from awkward._typing import TYPE_CHECKING, Any, DType, Self

np = NumpyMetadata.instance()

if TYPE_CHECKING:
    from numpy.typing import DTypeLike


class PlaceholderArray(MaterializableArray):
    def __init__(
        self,
        nplike: NumpyLike,
        shape: tuple[ShapeItem, ...],
        dtype: DType,
        buffer_key: str | None = None,
    ):
        self._nplike = nplike
        self._shape = tuple(dim if dim is unknown_length else int(dim) for dim in shape)
        self._dtype = np.dtype(dtype)
        self._buffer_key = buffer_key

    @property
    def buffer_key(self) -> str | None:
        return self._buffer_key

    @property
    def dtype(self) -> DType:
        return self._dtype

    @property
    def shape(self) -> tuple[ShapeItem, ...]:
        return self._shape

    @property
    def ndim(self) -> int:
        return len(self._shape)

    @property
    def size(self) -> ShapeItem:
        size: ShapeItem = 1
        for item in self._shape:
            size *= item
        return size

    @property
    def nbytes(self) -> int:
        return 0

    def materialize(self):
        msg = f"{self} should never have been encountered."
        if self.buffer_key:
            msg += (
                f" Awkward Array tried to access a buffer at '{self.buffer_key}', "
                "but it exists only as a placeholder."
            )
        msg += (
            " This is unexpected behavior — please open an issue at "
            "https://github.com/scikit-hep/awkward/issues with a minimal example."
        )
        raise RuntimeError(msg)

    @property
    def strides(self) -> tuple[ShapeItem, ...]:
        out: tuple[ShapeItem, ...] = (self._dtype.itemsize,)
        for item in reversed(self._shape):
            out = (item * out[0], *out)
        return out

    @property
    def T(self):
        return type(self)(self._nplike, self._shape[::-1], self._dtype)

    def view(self, dtype: DTypeLike) -> Self:
        dtype = np.dtype(dtype)
        if len(self._shape) >= 1:
            last, remainder = divmod(
                self._shape[-1] * self._dtype.itemsize, dtype.itemsize
            )
            if remainder is not unknown_length and remainder != 0:
                raise ValueError(
                    "new size of array with larger dtype must be a "
                    "divisor of the total size in bytes (of the last axis of the array)"
                )
            shape = (*self._shape[:-1], last)
        else:
            shape = self._shape
        return type(self)(self._nplike, shape, dtype, self._buffer_key)

    def __repr__(self):
        dtype = repr(self._dtype)
        if self.shape is None:
            shape = ""
        else:
            shape = f", shape={self._shape!r}"
        return f"PlaceholderArray({dtype}{shape})"

    def __getitem__(self, index):
        # Typetracers permit slices that don't touch data or shapes
        if isinstance(index, slice):
            length = self._shape[0]

            # Unknown-length placeholders should not be sliced (as their shapes would be touched(
            if length is unknown_length:
                raise AssertionError(
                    "placeholder arrays that are sliced should have known shapes"
                )
            # Known-length placeholders *always* need a known shape
            elif (
                index.start is unknown_length
                or index.stop is unknown_length
                or index.step is unknown_length
            ):
                raise AssertionError(
                    "known-length placeholders should never encounter unknown lengths in slices"
                )
            else:
                start, stop, step = index.indices(length)
                new_length = max(
                    0, (stop - start + (step - (1 if step > 0 else -1))) // step
                )

            return type(self)(
                self._nplike, (new_length,), self._dtype, self._buffer_key
            )
        else:
            msg = f"{type(self).__name__} supports only trivial slices, not {type(index).__name__}"
            if self.buffer_key:
                msg += f"\n\nAwkward-array attempted to access a buffer at '{self.buffer_key}', but "
                msg += (
                    "it has been excluded during a pre-run phase (possibly by Dask). "
                )
                msg += "If this was supposed to happen automatically (e.g. you're using Dask), "
                msg += "please report it to the developers at: https://github.com/scikit-hep/awkward/issues"
            raise TypeError(msg)

    def tolist(self):
        self.materialize()

    def __setitem__(self, key, value):
        del key
        maybe_materialize(self, value)

    def __bool__(self):  # pylint: disable=E0304
        self.materialize()

    def __int__(self):
        self.materialize()

    def __index__(self):  # noqa: PLE0305 # pylint: disable=E0305
        self.materialize()

    def __len__(self) -> int:
        return int(self._shape[0])

    def __add__(self, other):
        maybe_materialize(self, other)

    def __and__(self, other):
        maybe_materialize(self, other)

    def __eq__(self, other):
        maybe_materialize(self, other)

    def __floordiv__(self, other):
        maybe_materialize(self, other)

    def __ge__(self, other):
        maybe_materialize(self, other)

    def __gt__(self, other):
        maybe_materialize(self, other)

    def __invert__(self):
        self.materialize()

    def __le__(self, other):
        maybe_materialize(self, other)

    def __lt__(self, other):
        maybe_materialize(self, other)

    def __mul__(self, other):
        maybe_materialize(self, other)

    def __or__(self, other):
        maybe_materialize(self, other)

    def __sub__(self, other):
        maybe_materialize(self, other)

    def __truediv__(self, other):
        maybe_materialize(self, other)

    __iter__: None = None

    def __dlpack_device__(self):
        self.materialize()

    def __dlpack__(self, stream: Any = None):
        del stream
        self.materialize()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("to_tensorflow",)

np = NumpyMetadata.instance()


@high_level_function()
def to_tensorflow(array):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).

    Converts `array` (only ListOffsetArray, ListArray, RegularArray and NumpyArray data types supported)
    into a TensorFlow Tensor, if possible.

    If `array` contains any other data types (RecordArray for example) the function raises a TypeError.
    """

    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array)


def _impl(array):
    try:
        import tensorflow as tf
    except ImportError as err:
        raise ImportError(
            """to use ak.to_tensorflow, you must install the 'tensorflow' package with:

        pip install tensorflow
or
        conda install tensorflow"""
        ) from err

    # useful function that handles all possible input arrays
    array = ak.to_layout(array, allow_record=False)

    # get the device array is on
    ak_device = ak.backend(array)

    if ak_device not in ["cuda", "cpu"]:
        raise ValueError("""Only 'cpu' and 'cuda' backend conversions are allowed""")

    # convert to numpy or cupy if `array` on gpu
    try:
        backend_array = array.to_backend_array(allow_missing=False)
    except ValueError as err:
        raise TypeError(
            "Only arrays containing equal-length lists of numbers can be converted into a TensorFlow Tensor"
        ) from err

    if ak_device == "cpu":
        device = "CPU:0"
    else:
        id = backend_array.data.device.id
        device = "GPU:" + str(id)

    with tf.device(device):
        # check if cupy or numpy
        if isinstance(backend_array, np.ndarray):
            # convert numpy to a tensorflow tensor
            tensor = tf.convert_to_tensor(backend_array, dtype=tf.float64)
        else:
            # cupy -> tensorflow tensor
            tensor = tf.experimental.dlpack.from_dlpack(backend_array.toDlpack())

        return tensor



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import json

import pyarrow

from .extn_types import (
    AwkwardArrowArray,
    AwkwardArrowType,
    get_field_option,
    to_awkwardarrow_storage_types,
)

AWKWARD_INFO_KEY = b"awkward_array_metadata"  # metadata field in Table schema


def convert_awkward_arrow_table_to_native(aatable: pyarrow.Table) -> pyarrow.Table:
    """
    aatable: A pyarrow Table created with extensionarray=True
    returns: A pyarrow Table without extensionsarrays, but
      with `awkward_array_metadata` in the schema's metadata that can be used to
      convert the resulting table back into one with extensionarrays.
    """
    new_fields = []
    metadata = {}  # metadata for table column types
    for aacol_field in aatable.schema:
        metadata[aacol_field.name] = collect_ak_arr_type_metadata(aacol_field)
        new_field = awkward_arrow_field_to_native(aacol_field)
        new_fields.append(new_field)
    metadata_serial = json.dumps(metadata).encode(errors="surrogatescape")
    if aatable.schema.metadata is None:
        new_metadata = {}
    else:
        new_metadata = aatable.schema.metadata.copy()
    new_metadata[AWKWARD_INFO_KEY] = metadata_serial
    new_schema = pyarrow.schema(new_fields, metadata=new_metadata)

    return replace_schema(aatable, new_schema)


def convert_native_arrow_table_to_awkward(table: pyarrow.Table) -> pyarrow.Table:
    """
    table: A pyarrow Table converted with convert_awkward_arrow_table_to_native
    returns: A pyarrow Table without extensionsarrays, but
      with `awkward_array_metadata` in the schema's metadata that can be used to
      convert the resulting table back into one with extensionarrays.
    """
    if table.schema.metadata is None or AWKWARD_INFO_KEY not in table.schema.metadata:
        return table  # Prior versions don't include metadata here
    new_fields = []
    metadata = json.loads(
        table.schema.metadata[AWKWARD_INFO_KEY].decode(errors="surrogatescape")
    )
    for aacol_field in table.schema:
        if aacol_field.name not in metadata:
            raise ValueError(
                f"Awkward metadata in Arrow table does not have info for column {aacol_field.name}"
            )
        new_fields.append(
            native_arrow_field_to_akarraytype(aacol_field, metadata[aacol_field.name])
        )
    new_metadata = table.schema.metadata.copy()
    del new_metadata[AWKWARD_INFO_KEY]
    new_schema = pyarrow.schema(new_fields, metadata=new_metadata)
    # return table.cast(new_schema)  # Similar (same even?) results
    return replace_schema(table, new_schema)


def collect_ak_arr_type_metadata(aafield: pyarrow.Field) -> dict | list | None:
    """
    Given a Field, collect ArrowExtensionArray metadata as an object.
    If that field holds more ArrowExtensionArray types, a "subfield_metadata"
    property is added that holds a list of metadata objects for the sub-fields.
    This recurses down the whole type structure.
    """
    typ = aafield.type

    option_type = get_field_option(aafield, b"option_type")

    if not isinstance(typ, AwkwardArrowType):
        return None  # Not expected to reach here
    subfields = _fields_of_strg_type(typ.storage_type)
    metadata = typ._metadata_as_dict()
    metadata["field_name"] = aafield.name
    if option_type is False:
        metadata["option_type"] = "False"

    if len(subfields) == 0:
        # Simple type
        return metadata
    # Compound type
    subfield_metadata_list = []
    for ak_field in subfields:
        subfield_metadata_list.append(
            collect_ak_arr_type_metadata(ak_field)  # Recurse
        )
    metadata["subfield_metadata"] = subfield_metadata_list
    return metadata


def awkward_arrow_field_to_native(aafield: pyarrow.Field) -> pyarrow.Field:
    """
    Given a Field with ArrowExtensionArray type, returns a corresponding
    field with only Arrow builtin, or storage, types. Metadata is removed.
    """
    typ = aafield.type
    if not isinstance(typ, AwkwardArrowType):
        # Not expected to reach this. Maybe throw ValueError?
        return aafield

    fields = _fields_of_strg_type(typ.storage_type)
    if len(fields) == 0:
        # We have a simple type wrapped in AwkwardArrowType.
        new_field = pyarrow.field(
            aafield.name, type=typ.storage_type, nullable=aafield.nullable
        )
        return new_field

    # We have nested types
    native_fields = [
        awkward_arrow_field_to_native(field)  # Recurse
        for field in fields
    ]
    native_type = _make_pyarrow_type_like(typ, native_fields)
    new_field = pyarrow.field(aafield.name, type=native_type, nullable=aafield.nullable)
    return new_field


def native_arrow_field_to_akarraytype(
    ntv_field: pyarrow.Field, metadata: dict
) -> pyarrow.Field:
    if isinstance(ntv_field, AwkwardArrowType):
        raise ValueError(f"field {ntv_field} is already an AwkwardArrowType")
    storage_type = ntv_field.type
    fields = _fields_of_strg_type(storage_type)
    if len(fields) > 0:
        # We need to replace storage_type with one that contains AwkwardArrowTypes.
        sub_meta = metadata["subfield_metadata"]
        awkwardized_fields = None  # Temporary
        if len(sub_meta) == len(fields):
            awkwardized_fields = [
                native_arrow_field_to_akarraytype(field, meta)  # Recurse
                for field, meta in zip(
                    fields, metadata["subfield_metadata"], strict=True
                )
            ]
        elif len(fields) < len(sub_meta):
            # If a user has read a partial column, we can have fewer Arrow fields than the original.
            sub_meta_dict = {sm["field_name"]: sm for sm in sub_meta}
            awkwardized_fields = []
            for field in fields:
                if field.name in sub_meta_dict:
                    awkwardized_fields.append(
                        native_arrow_field_to_akarraytype(
                            field, sub_meta_dict[field.name]
                        )
                    )
                else:
                    raise ValueError(
                        f"Cannot find Awkward metadata for sub-field {field.name}"
                    )
        else:
            raise ValueError(
                f"Not enough fields in Awkward metadata. Have {len(sub_meta)} need at least {len(fields)}."
            )
        storage_type = _make_pyarrow_type_like(storage_type, awkwardized_fields)

    ak_type = AwkwardArrowType._from_metadata_object(storage_type, metadata)

    option_str = metadata.get("option_type") if metadata else None

    # Determine nullability: if option_type is 'False', force nullable to False
    nullable = False if option_str == "False" else ntv_field.nullable

    return pyarrow.field(
        ntv_field.name,
        type=ak_type,
        nullable=nullable,
    )


def _fields_of_strg_type(typ: pyarrow.Type) -> list[pyarrow.Field]:
    if isinstance(typ, pyarrow.lib.DictionaryType):
        return [
            pyarrow.field("value", typ.value_type)
        ]  # Wrap in a field for consistency
    elif typ.num_fields == 0:
        return []
    elif not hasattr(typ, "field"):
        # Old versions of pyarrow have this quirk.
        if hasattr(typ, "value_field"):
            return [typ.value_field]
        elif hasattr(typ, "__iter__"):
            return list(typ)
        raise TypeError(f"Cannot handle arrow type {typ}")
    else:
        return [typ.field(i) for i in range(typ.num_fields)]


def _make_pyarrow_type_like(
    typ: pyarrow.Type, fields: list[pyarrow.Field]
) -> pyarrow.Type:
    storage_type = to_awkwardarrow_storage_types(typ)[1]
    if isinstance(storage_type, pyarrow.lib.DictionaryType):
        return pyarrow.dictionary(storage_type.index_type, fields[0].type)
    elif isinstance(storage_type, pyarrow.lib.FixedSizeListType):
        return pyarrow.list_(fields[0], storage_type.list_size)
    elif isinstance(storage_type, pyarrow.lib.ListType):
        return pyarrow.list_(fields[0])
    elif isinstance(storage_type, pyarrow.lib.LargeListType):
        return pyarrow.large_list(fields[0])
    elif isinstance(storage_type, pyarrow.lib.MapType):
        # return pyarrow.map_(storage_type.index_type, fields[0])
        raise NotImplementedError("pyarrow MapType is not supported by Awkward")
    elif isinstance(storage_type, pyarrow.lib.StructType):
        return pyarrow.struct(fields)
    elif isinstance(storage_type, pyarrow.lib.UnionType):
        return pyarrow.union(fields, storage_type.mode, storage_type.type_codes)
    elif (
        isinstance(storage_type, pyarrow.lib.DataType) and storage_type.num_fields == 0
    ):
        # Catch-all for primitive types, nulltype, string types, FixedSizeBinaryType
        return storage_type
    raise NotImplementedError(f"Type {typ} is not handled for conversion.")


def replace_schema(table: pyarrow.Table, new_schema: pyarrow.Schema) -> pyarrow.Table:
    """
    This function is like `pyarrow.Table.cast()` except it only works if the
    new schema uses the same storage types and storage geometries as the original.
    It explicitly will not convert one primitive type to another.
    """
    new_batches = []
    for batch in table.to_batches():
        columns = []
        for col, new_field in zip(batch.columns, new_schema, strict=True):
            columns.append(array_with_replacement_type(col, new_field.type))
        new_batches.append(
            pyarrow.RecordBatch.from_arrays(arrays=columns, schema=new_schema)
        )
    return pyarrow.Table.from_batches(new_batches, schema=new_schema)


def array_with_replacement_type(
    orig_array: pyarrow.Array, new_type: pyarrow.Type
) -> pyarrow.Array:
    """
    Creates a new array with a different type.
    Either pyarrow native -> ExtensionArray or vice-versa.
    """
    children_orig = _get_children(orig_array)
    native_type = to_awkwardarrow_storage_types(new_type)[1]
    new_fields = _fields_of_strg_type(native_type)
    if len(new_fields) != len(children_orig):
        raise AssertionError(
            f"Number of children: {len(children_orig) =} != {len(new_fields) =}"
        )
    children_new = [
        array_with_replacement_type(child, new_child_type.type)
        for child, new_child_type in zip(children_orig, new_fields, strict=True)
    ]
    own_buffers = orig_array.buffers()[: orig_array.type.num_buffers]
    if isinstance(native_type, pyarrow.lib.DictionaryType):
        # The following works with newer pyarrow versions, but not 7.0:
        # native_dict = pyarrow.DictionaryArray.from_buffers(
        #     type=native_type,
        #     length=len(orig_array),
        #     buffers=own_buffers,
        #     dictionary=children_new[0],
        #     null_count=orig_array.null_count,
        #     offset=orig_array.offset,
        # )
        if isinstance(orig_array, AwkwardArrowArray):
            native_orig = orig_array.storage
        else:
            native_orig = orig_array
        native_dict = pyarrow.DictionaryArray.from_arrays(
            indices=native_orig.indices,
            dictionary=children_new[0],
            mask=own_buffers[0],
            safe=False,
        )
        if isinstance(new_type, pyarrow.ExtensionType):
            return AwkwardArrowArray.from_storage(new_type, native_dict)
        else:
            return native_dict
    else:
        return pyarrow.Array.from_buffers(
            type=new_type,
            length=len(orig_array),
            buffers=own_buffers,
            null_count=orig_array.null_count,
            offset=orig_array.offset,
            children=children_new,
        )


def _get_children(array: pyarrow.Array) -> list[pyarrow.Array]:
    """
    Different types of pyarrow Arrays have different ways to
    access their "children." It helps to unify these.
    """
    arrow_type = to_awkwardarrow_storage_types(array.type)[1]
    if isinstance(array, AwkwardArrowArray):
        array = array.storage

    if isinstance(arrow_type, pyarrow.lib.DictionaryType):
        return [array.dictionary]
    if arrow_type.num_fields == 0:
        return []
    if hasattr(array, "field"):
        return [array.field(idx) for idx in range(arrow_type.num_fields)]
    if hasattr(array, "values"):
        return [array.values]
    raise NotImplementedError(f"Cannot get children of arrow type {arrow_type}")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from collections.abc import Callable, Iterator

import awkward as ak
from awkward._meta.regularmeta import RegularMeta
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._nplikes.shape import unknown_length
from awkward._regularize import is_integer
from awkward._typing import Any, DType, Self, final
from awkward._util import UNSET
from awkward.forms.form import Form, _SpecifierMatcher

__all__ = ("RegularForm",)

np = NumpyMetadata.instance()


@final
class RegularForm(RegularMeta[Form], Form):
    _content: Form

    def __init__(self, content, size, *, parameters=None, form_key=None):
        if not isinstance(content, Form):
            raise TypeError(
                f"{type(self).__name__} all 'contents' must be Form subclasses, not {content!r}"
            )
        if not (size is unknown_length or (is_integer(size) and size >= 0)):
            raise TypeError(
                f"{type(self).__name__} 'size' must be a non-negative int or None, not {size!r}"
            )

        self._content = content
        self._size = size
        self._init(parameters=parameters, form_key=form_key)

    @property
    def content(self):
        return self._content

    @property
    def size(self):
        return self._size

    def copy(self, content=UNSET, size=UNSET, *, parameters=UNSET, form_key=UNSET):
        return RegularForm(
            self._content if content is UNSET else content,
            self._size if size is UNSET else size,
            parameters=self._parameters if parameters is UNSET else parameters,
            form_key=self._form_key if form_key is UNSET else form_key,
        )

    @classmethod
    def simplified(cls, content, size, *, parameters=None, form_key=None):
        return cls(content, size, parameters=parameters, form_key=form_key)

    def __repr__(self):
        args = [repr(self._content), repr(self._size), *self._repr_args()]
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def _to_dict_part(self, verbose, toplevel):
        return self._to_dict_extra(
            {
                "class": "RegularArray",
                "size": None if self._size is unknown_length else self._size,
                "content": self._content._to_dict_part(verbose, toplevel=False),
            },
            verbose,
        )

    @property
    def type(self):
        return ak.types.RegularType(
            self._content.type,
            self._size,
            parameters=self._parameters,
        )

    def _columns(self, path, output, list_indicator):
        if (
            self.parameter("__array__") not in ("string", "bytestring")
            and list_indicator is not None
        ):
            path = (*path, list_indicator)
        self._content._columns(path, output, list_indicator)

    def _prune_columns(self, is_inside_record_or_union: bool) -> Self | None:
        next_content = self._content._prune_columns(is_inside_record_or_union)
        if next_content is None:
            return None
        else:
            return self.copy(content=next_content)

    def _select_columns(self, match_specifier: _SpecifierMatcher) -> Self:
        return self.copy(content=self._content._select_columns(match_specifier))

    def _column_types(self):
        if self.parameter("__array__") in ("string", "bytestring"):
            return ("string",)
        else:
            return self._content._column_types()

    def __setstate__(self, state):
        if isinstance(state, dict):
            # read data pickled in Awkward 2.x
            self.__dict__.update(state)
        else:
            # read data pickled in Awkward 1.x

            # https://github.com/scikit-hep/awkward/blob/main-v1/src/python/forms.cpp#L692-L698
            _has_identities, parameters, form_key, content, size = state

            if form_key is not None:
                form_key = "part0-" + form_key  # only the first partition

            self.__init__(content, size, parameters=parameters, form_key=form_key)

    def _expected_from_buffers(
        self, getkey: Callable[[Form, str], str], recursive: bool
    ) -> Iterator[tuple[str, DType]]:
        if recursive:
            yield from self._content._expected_from_buffers(getkey, recursive)

    def _is_equal_to(self, other: Any, all_parameters: bool, form_key: bool) -> bool:
        return (
            self._is_equal_to_generic(other, all_parameters, form_key)
            and (self._size == other._size)
            and self._content._is_equal_to(other._content, all_parameters, form_key)
        )
